"""
    Module with all tracking related entities 
"""

from typing import Type
from datetime import datetime

import dateutil.parser
from pypaypal.entities.base import T, PayPalEntity, ResponseType, ActionLink

class Tracker(PayPalEntity):
    """
        Object representation of a paypal tracker
    """
    def __init__(self, transaction_id: str, tracking_number: int, status: str, carrier: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.status = status
        self.carrier = carrier
        self.transaction_id = transaction_id
        self.tracking_number = tracking_number
        self.quantity = self._json_response.get('quantity') or kwargs.get('quantity')
        self.notify_buyer = self._json_response.get('notify_buyer') or kwargs.get('notify_buyer')
        self._shipment_date = self._json_response.get('shipment_date') or kwargs.get('shipment_date')
        self._last_updated_time = self._json_response.get('last_updated_time') or kwargs.get('last_updated_time')
        self.carrier_name_other = self._json_response.get('carrier_name_other') or kwargs.get('carrier_name_other')
        self.postage_payment_id = self._json_response.get('postage_payment_id') or kwargs.get('postage_payment_id')
        self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]
        self.tracking_number_validated = self._json_response.get('tracking_number_validated') or kwargs.get('tracking_number_validated')

    @property
    def shipment_date(self) -> datetime:
        try:
            return datetime.strptime(self._shipment_date, '%Y-%m-%d') if self._shipment_date else None
        except:
            return None

    @property
    def last_updated_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._last_updated_time) if self._last_updated_time else None
        except:
            return None

    @property
    def read_link(self) -> ActionLink:
        """Retrieves a link to read this entity details.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'self', self.links), None)

    @property
    def update_link(self) -> ActionLink:
        """Retrieves the update link for this entity
        
        Returns:
            ActionLink -- The link for requesting an update in the API.
        """
        return next(filter(lambda x: x.rel == 'replace', self.links), None)

    @property
    def create_link(self) -> ActionLink:
        """Retrieves the link to create a tracker or add a tracker to an order
        
        Returns:
            ActionLink -- The link for requesting a create operation in the API.
        """
        return next(filter(lambda x: x.rel == 'create', self.links), None)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(
            json_data['transaction_id'], json_data['tracking_number'], json_data['status'],
            json_data['carrier'], json_response= json_data, response_type = response_type
        )



"""
    Sync (Transaction) resource clients for the paypal resource group.

    Resource docs & Reference: https://developer.paypal.com/docs/api/sync/v1/

    It seems that paypal renamed this endpoint in the docs to: https://developer.paypal.com/docs/api/transaction-search/v1/
"""

import json

from enum import Enum
from datetime import datetime
from typing import TypeVar, Set

from pypaypal.clients.base import ClientBase, ActionLink

from pypaypal.http import ( 
    parse_url,
    PayPalSession,
    LEGACY_LIVE_API_BASE_URL, 
    LEGACY_SANDBOX_API_BASE_URL 
)

from pypaypal.entities.base import ( 
    Money, 
    DateRange,
    AmountRange,
    ResponseType,
    PaypalApiResponse
)

from pypaypal.entities.sync import TransactionStatus, TransactionResponse

"""
    Base Resource Live URL
"""
_LIVE_RESOURCE_BASE_URL = parse_url(LEGACY_LIVE_API_BASE_URL, 'reporting', 'transactions')

"""
    Base Resource Sandbox URL
"""
_SANDBOX_RESOURCE_BASE_URL = parse_url(LEGACY_SANDBOX_API_BASE_URL, 'reporting', 'transactions')

T = TypeVar('T', bound = 'SyncClient')

_TRANSACTION_STRING_VALUES = [ 
    'transaction_info', 'payer_info', 'shipping_info', 'auction_info', 
    'cart_info', 'incentive_info', 'store_info', 'all' 
]

class TransactionField(Enum):
    #  The transaction information. 
    TRANSACTION_INFO = 1
    #  The payer information. 
    PAYER_INFO = 2
    #  The shipping information.
    SHIPPING_INFO = 3
    #  The auction information.
    AUCTION_INFO = 4
    #  The cart information.
    CART_INFO = 5
    #  An array of incentive detail objects.
    INCENTIVE_INFO = 6
    #  The store information.
    STORE_INFO = 7
    # All
    ALL = 8

    def as_parameter(self) -> str:
        return _TRANSACTION_STRING_VALUES[self.value - 1]

class PaymentInsrtumentType(Enum):
    # To query for a debit card transaction with a corresponding value.
    DEBITCARD = 1
    # To query for a direct credit card transaction with a corresponding value.
    CREDITCARD = 2

class SyncClient(ClientBase):
    """Sync (Transactio reporting) resource client
    """
    def __init__(self, base_url: str, session: PayPalSession):
        super().__init__(base_url, session)
    
    def list_transactions(
        self, *, page: int = 1, page_size: int = 100,  fields: Set[TransactionField] = { TransactionField.TRANSACTION_INFO },
        transaction_id: str = None, transaction_type: str = None, transaction_status: TransactionStatus = None, 
        transaction_amount: AmountRange = None, date: DateRange = None, payment_instrument_type: PaymentInsrtumentType = None, 
        store_id: str = None, terminal_id: str = None, balance_affecting_records_only: bool = True
    ) -> PaypalApiResponse[TransactionResponse]:
        """[summary]
        
        Keyword Arguments:
            page {int} -- Curent page (default: {1})
            page_size {int} -- page size max 500 (default: {100})
            fields {Set[TransactionField]} -- fields to include in the report (default: {{ TransactionField.TRANSACTION_INFO }})
            transaction_id {str} -- transaction ids (note: they are not unique) (default: {None})
            transaction_type {str} -- transaction event code (default: {None})
            transaction_status {TransactionStatus} -- transaction status for the query (default: {None})
            transaction_amount {AmountRange} -- transaction amount range filter (default: {None})
            date {DateRange} -- transaction date range filter (default: {None})
            payment_instrument_type {PaymentInsrtumentType} -- creditcard or debitcard payment (None for both) (default: {None})
            store_id {str} --  Filters the transactions in the response by a store ID.  (default: {None})
            terminal_id {str} -- Filters the transactions in the response by a terminal ID.  (default: {None})
            balance_affecting_records_only {bool} --  flag to include only balance-impacting or all transactions (default: {True})
        
        Returns:
            PaypalApiResponse[TransactionResponse] -- Page with all the desired transaction info
        """
        parameters = {
            'page': page, 'store_id': store_id, 'page_size': page_size,
            'terminal_id': terminal_id, 'transaction_id': transaction_id,
            'transaction_type': transaction_type, 'transaction_status': transaction_status,            
            'balance_affecting_records_only': balance_affecting_records_only
        }

        parameters = { k:v for k,v in parameters.items() if v != None }

        if date:
            parameters['end_date'] = date.end
            parameters['start_date'] = date.start
            
        if payment_instrument_type:
            parameters['payment_instrument_type'] = payment_instrument_type.name
        
        if transaction_amount:
            # Both currency codes should be equal in this request
            parameters['transaction_currency'] = transaction_amount.lower_amount.currency_code            
            parameters['transaction_amount'] = f'{transaction_amount.lower_amount} TO {transaction_amount.upper_amount}'\
                .replace('.', '')
            
        if fields:
            f = ','.join([x.as_parameter() for x in fields])
            if TransactionField.ALL in fields or len(fields) == 7:
                f = TransactionField.ALL.as_parameter()
            parameters['fields'] = f

        api_response = self._session.get(self._base_url, parameters)

        if api_response.status_code // 100 != 2:
            return PaypalApiResponse.error(api_response)

        return PaypalApiResponse.success(api_response, TransactionResponse.serialize_from_json(api_response.json()))

    @classmethod
    def for_session(cls: T, session: PayPalSession) -> T:
        """Creates a client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of Dispute client with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(base_url, session)


"""
    Module for Webhook related entities.
"""

from enum import Enum
from datetime import datetime
from typing import Type, List

import dateutil.parser

from pypaypal.entities.base import (
    T,
    ActionLink,
    ResponseType,    
    PayPalEntity
)

class AnchorType(Enum):
    """Webhook anchor type for filtering
    """
    ACCOUNT = 1
    APPLICATION = 2

class SignatureVerificationStatus(Enum):
    SUCCESS = 1
    FAILURE = 2

class ResourceVersion(PayPalEntity):
    """Resource version object representation
    """

    def __init__(self, resource_version: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.resource_version = resource_version
    
    @classmethod
    def create(cls, resource_version: str) -> 'ResourceVersion':
        return cls(resource_version)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(**json_data, json_response = json_data, response_type = response_type)

class EventType(PayPalEntity):
    """Event type obj representation
    """

    _ARRAY_TYPES = { 'resource_version': ResourceVersion }

    def __init__(
            self, name: str = None, description: str = None, status: str = None, 
            resource_version: List[ResourceVersion] = [], **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.name = name
        self.description = description
        self.status = status
        self.resource_version = resource_version or []
    
    @classmethod
    def create(cls, name: str, description: str, status: str, resource_version: List[ResourceVersion]) -> 'EventType':
        return cls(name, description, status, resource_version)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = super()._build_args(json_data, entity_types = dict(), array_types = cls._ARRAY_TYPES)
        return cls(**args, json_response = json_data, response_type = response_type)

class Webhook(PayPalEntity):
    """Webhook obj representation
    """

    _ARRAY_TYPES = { 'event_types': EventType }

    def __init__(self, webhook_id: str = None, url: str = None, event_types: List[EventType] = [], **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.url = url
        self.webhook_id = webhook_id
        self.event_types = event_types or []
        self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]

    def to_dict(self) -> dict:
        ret = super().to_dict()
        if ret['webhook_id']:
            ret['id'] = ret.pop('webhook_id')
        return ret

    @classmethod
    def create(cls, url: str, event_types: List[EventType]) -> 'Webhook':
        return cls(None, url, event_types)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = super()._build_args(json_data, entity_types = dict(), array_types = cls._ARRAY_TYPES)
        return cls(**args, json_response = json_data, response_type = response_type)
    
class ResourceAmountDetails(PayPalEntity):
    """Resource Details obj representation
    """

    def __init__(self, subtotal: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.subtotal = subtotal
    
    @classmethod
    def create(cls, subtotal: str) -> 'ResourceAmountDetails':
        return cls(subtotal)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:        
        return cls(**json_data, json_response = json_data, response_type = response_type)

class ResourceAmount(PayPalEntity):
    """Resource amount obj representation
    """

    _ENTITY_TYPES = { 'details': ResourceAmountDetails }

    def __init__(self, total: str = None, currency: str = None, details: ResourceAmountDetails = None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.total = total
        self.details = details
        self.currency = currency
    
    @classmethod
    def create(cls, total: str, currency: str, details: ResourceAmountDetails) -> 'ResourceAmount':
        return cls(total, currency, details)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:        
        args = super()._build_args(json_data, cls._ENTITY_TYPES)
        return cls(**args, json_response = json_data, response_type = response_type)

class Resource(PayPalEntity):
    """Resource obj representation
    """
    
    _ENTITY_TYPES = { 'amount': ResourceAmount }

    def __init__(self, resource_id: str = None, state: str = None, amount: ResourceAmount = None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.state = state
        self.amount = amount
        self.resource_id = resource_id
        self._update_time = self._json_response.get('update_time', kwargs.get('update_time'))
        self._create_time = self._json_response.get('create_time', kwargs.get('create_time'))
        self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]

    def to_dict(self) -> dict:
        ret = super().to_dict()
        if ret['resource_id']:
            ret['id'] = ret.pop('resource_id')
        return ret

    @property
    def update_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._update_time) if self._update_time else None
        except:
            return None

    @property
    def create_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._create_time) if self._create_time else None
        except:
            return None

    @classmethod
    def create(cls, resource_id: str, state: str, amount: ResourceAmount) -> 'Resource':
        return cls(resource_id, state, amount)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = super()._build_args(json_data, cls._ENTITY_TYPES)
        return cls(**args, json_response = json_data, response_type = response_type)

class WebhookEvent(PayPalEntity):
    """WebhookEvent obj representation
    """
    
    _ENTITY_TYPES = { 'resource': Resource }

    def __init__(
            self, event_id: str = None, resource_type: str = None, 
            event_version: str = None, event_type: str = None, summary: str = None, 
            resource_version: str = None, resource: Resource = None, **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.summary = summary
        self.event_id = event_id
        self.resource = resource
        self.event_type = event_type
        self.resource_type = resource_type
        self.event_version = event_version
        self.resource_version = resource_version
        self._create_time = self._json_response.get('create_time', kwargs.get('create_time'))
        self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]

    def to_dict(self) -> dict:
        ret = super().to_dict()
        if ret['event_id']:
            ret['id'] = ret.pop('event_id')
        return ret

    @property
    def create_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._create_time) if self._create_time else None
        except:
            return None

    @classmethod
    def create(cls, event_id: str = None, resource_type: str = None,
            event_version: str = None, event_type: str = None, summary: str = None,
            resource_version: str = None, resource: Resource = None) -> 'WebhookEvent':
        return cls(
            summary = summary, event_id = event_id, event_type = event_type,
            resource_type = resource_type, event_version = event_version,
            resource = resource, resource_version = resource_version
        )

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = super()._build_args(json_data, cls._ENTITY_TYPES)
        return cls(**args, json_response = json_data, response_type = response_type)

class WebhookSignature(PayPalEntity):
    """Webhook cert Signature obj representation
    """

    _ENTITY_TYPES = { 'webhook_event': WebhookEvent }

    def __init__(
            self, auth_algo: str = None, cert_url: str = None, 
            transmission_id: str = None, transmission_sig: str = None, 
            transmission_time: str = None, webhook_id: str = None, 
            webhook_event: WebhookEvent = None, **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.cert_url = cert_url
        self.auth_algo = auth_algo
        self.webhook_id = webhook_id
        self.webhook_event = webhook_event
        self.transmission_id = transmission_id
        self.transmission_sig = transmission_sig
        self.transmission_time = transmission_time
    
    @classmethod
    def create(
            cls, auth_algo: str, cert_url: str, transmission_id: str, 
            transmission_sig: str, transmission_time: str, webhook_id: str, 
            webhook_event: WebhookEvent
        ) -> 'WebhookSignature':
        return cls(
                auth_algo = auth_algo, cert_url = cert_url, 
                transmission_id = transmission_id, transmission_sig = transmission_sig, 
                transmission_time = transmission_time, webhook_id = webhook_id, 
                webhook_event = webhook_event 
            )

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = super()._build_args(json_data, cls._ENTITY_TYPES)
        return cls(**args, json_response = json_data, response_type = response_type)





"""Module with basic http constants & session handling.
"""
import urllib.parse

from enum import Enum
from abc import ABC, abstractmethod

from requests import ( 
    get, post, put, patch, delete 
)

from requests.auth import HTTPBasicAuth

from typing import NamedTuple
from datetime import datetime, timedelta

from pypaypal.errors import IdentityError, ExpiredSessionError

"""
    Live PayPal api base URL.

    This library is intended for paypal integration with the v2 api if possible.
"""
LIVE_API_BASE_URL = 'https://api.paypal.com/v2/'

"""
    Sandbox PayPal api base URL.

    This library is intended for paypal integration with the v2 api if possible.
"""
SANDBOX_API_BASE_URL = 'https://api.sandbox.paypal.com/v2/'

"""Live PayPal api base URL for v1 api requests.
"""
LEGACY_LIVE_API_BASE_URL = 'https://api.paypal.com/v1/'

"""Sandbox PayPal api base URL for v1 api requests.
"""
LEGACY_SANDBOX_API_BASE_URL = 'https://api.sandbox.paypal.com/v1/'

class AuthType(Enum):
    """
        Enumerated constants for the session type
    """
    BASIC = 1
    TOKEN = 2
    REFRESHABLE = 3

class SessionMode(Enum):
    """Session mode
    """
    LIVE = 1
    SANDBOX = 2

    def is_live(self) -> bool:
        """Boolean flag dictating if this instance is a live session mode or not
        
        Returns:
            bool -- true if live false if sandbox
        """
        return self == SessionMode.LIVE

class SessionStatus(Enum):
    ACTIVE = 1
    EXPIRED = 2
    DISPOSED = 3

class PayPalToken(NamedTuple):
    """Paypal access token wrapper.
    """
    scope: str
    access_token: str
    token_type: str
    app_id: str
    expires_in: int
    nonce: str
    requested_at: datetime

    def is_expired(self) -> bool:
        """
            Checks if the token is expired
        """
        return self.requested_at + timedelta(seconds = self.expires_in) < datetime.now()
    
    @classmethod
    def serialize(cls, json_data: dict):
        """Serializes a paypal json OAuthResponse into an instance
        """
        return cls(
            json_data['scope'], json_data['access_token'], json_data['token_type'], 
            json_data['app_id'], json_data['expires_in'], json_data['nonce'], datetime.now()
        )

def parse_url(base: str, *args: str) -> str:
    """Safely parses a web url
    """
    # Making sure the base ends with '/'
    base = base.strip('/') + '/'
    if len(args) > 1:
        arg = args[0].strip('/')
        return parse_url(urllib.parse.urljoin(base, arg), *args[1:])
    return urllib.parse.urljoin(base, args[0].strip('/'))

def _authenticate(client_id: str, secret:str, mode: SessionMode) -> PayPalToken:
    """Basic authentication that returns a PayPalToken
    
    Arguments:
        client_id {str} -- paypal client id
        secret {str} -- paypal client secret
        mode {SessionMode} -- Desired session mode (LIVE or SANDBOX)
    
    Raises:
        IdentityError: If the user fails to authenticate
    
    Returns:
        PayPalToken -- An immutable object with the token information
    """
    base = LEGACY_LIVE_API_BASE_URL if mode == SessionMode.LIVE else LEGACY_SANDBOX_API_BASE_URL
    url = parse_url(base, '/oauth2/token')
    body = { 'grant_type' : 'client_credentials' }

    headers = {        
        'Accept': 'application/json',
        'Content-Type': 'application/x-www-form-urlencoded'
    }

    response = post(url, body, None, auth=HTTPBasicAuth(client_id, secret), headers=headers)

    if response.status_code != 200:
        raise IdentityError(response)

    return PayPalToken.serialize(response.json())

class PayPalSession(ABC):
    """PayPal session abstraction
    """
    def __init__(self, auth_type: AuthType, session_mode: SessionMode, token: PayPalToken):
        """Constructor
        
        Arguments:
            auth_type {AuthType} -- The instance session type
            session_mode {SessionMode} -- The instance session mode
            token {PayPalToken} -- The instance initial access token
        """
        self._paypal_token = token
        self.auth_type = auth_type
        self.session_mode = session_mode
        self.status = SessionStatus.ACTIVE
    
    @abstractmethod
    def get(self, url: str, params:dict=None, **kwargs):
        """Secured get request
        
        Arguments:
            url {str} -- The request URL
        
        Keyword Arguments:
            params {dict} -- query string parameters (default: {None})
    
        Raises:
            ExpiredSessionError: If the session is in an invalid state

        Returns:
            An http response
        """
        pass

    @abstractmethod
    def post(self, url: str, body, **kwargs):
        """Secured post request
        
        Arguments:
            url {str} -- Request URL
            body {[type]} -- Request body
        """
        pass

    @abstractmethod
    def put(self, url: str, body, **kwargs):
        """Secured put request
        
        Arguments:
            url {str} -- Request URL
            body {[type]} -- Request body
        """
        pass

    @abstractmethod
    def patch(self, url: str, body, **kwargs):
        """Secured patch request
        
        Arguments:
            url {str} -- Request URL
            body {[type]} -- Request body
        """
        pass

    @abstractmethod
    def delete(self, url: str, **kwargs):
        """Secured patch request
        
        Arguments:
            url {str} -- Request URL
        """
        pass

    @abstractmethod
    def _dispose(self):
        """Disposes this instance
        """
        pass

    def _prepare_headers(self, token: PayPalToken, custom_headers:dict=None) -> dict:
        """Gets the headers to be used on every request
        
        Arguments:
            token {PayPalToken} -- access token
        
        Keyword Arguments:
            custom_headers {dict} -- custom headers (default: {None})
        
        Returns:
            dict -- a merged dictionary with basic and custom headers
        """
        headers = {
            'Content-Type': 'application/json',
            'Authorization': f'{token.token_type} {token.access_token}'
        }

        return { **headers, **custom_headers } if custom_headers else headers

    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        self._dispose()

class _OAuthSession(PayPalSession):
    """PayPal session obj for request with OAuthToken headers.

        A standard session with a finite lifespan,
        once the session is expired this instance will not be valid for further
        requests.
    """
    def __init__(self, session_mode: SessionMode, token: PayPalToken):
        super().__init__(AuthType.TOKEN, session_mode, token)
    
    def _check_token(self) -> PayPalToken:
        if self.status == SessionStatus.ACTIVE and (self._paypal_token == None or self._paypal_token.is_expired()):
            self.status = SessionStatus.EXPIRED
        
        if self.status != SessionStatus.ACTIVE:        
            raise ExpiredSessionError(self)

        return self._paypal_token

    def get(self, url: str, params:dict=None, **kwargs):
        kwargs['headers'] = self._prepare_headers(self._check_token(), kwargs.get('headers'))
        return get(url, params, **kwargs)

    def post(self, url: str, body, **kwargs):
        kwargs['headers'] = self._prepare_headers(self._check_token(), kwargs.get('headers'))
        return post(url, body, None, **kwargs)

    def put(self, url: str, body, **kwargs):
        kwargs['headers'] = self._prepare_headers(self._check_token(), kwargs.get('headers'))
        return put(url, body, **kwargs)

    def patch(self, url: str, body, **kwargs):
        kwargs['headers'] = self._prepare_headers(self._check_token(), kwargs.get('headers'))
        return patch(url, body, **kwargs)

    def delete(self, url: str, **kwargs):
        kwargs['headers'] = self._prepare_headers(self._check_token(), kwargs.get('headers'))
        return delete(url, **kwargs)

    def _dispose(self):
        self._paypal_token = None
        self.status = SessionStatus.DISPOSED

    def __repr__(self):
        return f'_OAuthSession(session_mode={self.session_mode}, status={self.status})'

    def __str__(self):
        return f'_OAuthSession(session_mode={self.session_mode}, status={self.status})'
    
class _BasicAuthSession(PayPalSession):
    """PayPal session obj for request with Basic Authorization.

        A session with an indefinite lifespan,
        all requests will use basic authorization which means that
        the client & secret will always travel through the network.
    """
    def __init__(self, session_mode: SessionMode, token: PayPalToken, client: str, secret: str):
        super().__init__(AuthType.BASIC, session_mode, token)
        self._client = client
        self._secret = secret
    
    def _check_token(self) -> PayPalToken:
        if  self._paypal_token != None and self._paypal_token.is_expired():
            self._paypal_token = None
        
        if self.status != SessionStatus.ACTIVE:
            raise ExpiredSessionError(self)

        return self._paypal_token

    def get(self, url: str, params:dict=None, **kwargs):
        token = self._check_token()        
        if token:
            # If there's a valid token we might as well use it
            kwargs['headers'] = self._prepare_headers(token, kwargs.get('headers'))
        else:
            kwargs['auth'] = HTTPBasicAuth(self._client, self._secret)
        
        response = get(url, params, **kwargs)        
        if response.status_code == 401:
            raise IdentityError(response)
        return response

    def post(self, url: str, body, **kwargs):
        token = self._check_token()        
        if token:
            # If there's a valid token we might as well use it
            kwargs['headers'] = self._prepare_headers(token, kwargs.get('headers'))
        else:
            kwargs['auth'] = HTTPBasicAuth(self._client, self._secret)

        response = post(url, body, None, **kwargs)        
        if response.status_code == 401:
            raise IdentityError(response)
        return response

    def put(self, url: str, body, **kwargs):
        token = self._check_token()        
        if token:
            # If there's a valid token we might as well use it
            kwargs['headers'] = self._prepare_headers(token, kwargs.get('headers'))
        else:
            kwargs['auth'] = HTTPBasicAuth(self._client, self._secret)
        
        response = put(url, body, **kwargs)
        if response.status_code == 401:
            raise IdentityError(response)
        return response

    def patch(self, url: str, body, **kwargs):
        token = self._check_token()        
        if token:
            # If there's a valid token we might as well use it
            kwargs['headers'] = self._prepare_headers(token, kwargs.get('headers'))
        else:
            kwargs['auth'] = HTTPBasicAuth(self._client, self._secret)
        
        response = patch(url, body, **kwargs)
        if response.status_code == 401:
            raise IdentityError(response)
        return response

    def delete(self, url: str, **kwargs):
        token = self._check_token()        
        if token:
            # If there's a valid token we might as well use it
            kwargs['headers'] = self._prepare_headers(token, kwargs.get('headers'))
        else:
            kwargs['auth'] = HTTPBasicAuth(self._client, self._secret)

        response = delete(url, **kwargs)
        if response.status_code == 401:
            raise IdentityError(response)
        return response

    def _dispose(self):
        self._client = None
        self._secret = None
        self._paypal_token = None
        self.status = SessionStatus.DISPOSED
    
    def __repr__(self):
        return f'_BasicAuthSession(session_mode={self.session_mode}, status={self.status}, client={self._client})'

    def __str__(self):
        return f'_BasicAuthSession(session_mode={self.session_mode}, status={self.status}, client={self._client})'

class _RefreshableSession(PayPalSession):
    """PayPal session obj for request with with OAuthToken headers and Basic Authorization.

        A mixed session using the standard _OAuthSession approach for regular requests
        and a _BasicAuthSession request to perform a token refresh. This gives the session
        and indefinite lifespan and a limited number of Basic Auth requests with a client &
        secret network roundtrips.
        
        This session can receive flags to limit the refresh count.
    """
    def __init__(self, session_mode: SessionMode, token: PayPalToken, client:str, secret: str, refresh_limit:int=None):
        super().__init__(AuthType.REFRESHABLE, session_mode, token)
        self._client = client
        self._secret = secret
        self._refresh_limit = refresh_limit
    
    def _check_token(self) -> PayPalToken:
        if self.status == SessionStatus.ACTIVE and (self._paypal_token == None or self._paypal_token.is_expired()):
            self.status = SessionStatus.EXPIRED
        
        if self.status == SessionStatus.EXPIRED and (self._refresh_limit == None or self._refresh_limit > 0):
            self._refresh_limit-=1
            self._paypal_token = _authenticate(self._client, self._secret, self.session_mode)
            self.status = SessionStatus.ACTIVE

        if self.status != SessionStatus.ACTIVE:
            raise ExpiredSessionError(self)

        return self._paypal_token

    def get(self, url: str, params:dict=None, **kwargs):
        kwargs['headers'] = self._prepare_headers(self._check_token(), kwargs.get('headers'))
        return get(url, params, **kwargs)

    def post(self, url: str, body, **kwargs):
        kwargs['headers'] = self._prepare_headers(self._check_token(), kwargs.get('headers'))
        return post(url, body, None, **kwargs)

    def put(self, url: str, body, **kwargs):
        kwargs['headers'] = self._prepare_headers(self._check_token(), kwargs.get('headers'))
        return put(url, body, **kwargs)

    def patch(self, url: str, body, **kwargs):
        kwargs['headers'] = self._prepare_headers(self._check_token(), kwargs.get('headers'))
        return patch(url, body, **kwargs)

    def delete(self, url: str, **kwargs):
        kwargs['headers'] = self._prepare_headers(self._check_token(), kwargs.get('headers'))
        return delete(url, **kwargs)    

    def _dispose(self):
        self._client = None
        self._secret = None
        self._refresh_limit = 0
        self._paypal_token = None
        self.status = SessionStatus.DISPOSED

    def __repr__(self):
        return f'_RefreshableSession(session_mode={self.session_mode}, status={self.status}, client={self._client})'

    def __str__(self):
        return f'_RefreshableSession(session_mode={self.session_mode}, status={self.status}, client={self._client})'

def session_from_token(token: PayPalToken, mode: SessionMode) -> PayPalSession:
    """Creates a session from a given token
    
    Arguments:
        token {PayPalToken} -- A valid paypal token instance 
        mode {SessionMode} -- Desired session mode (LIVE or SANDBOX)
    
    Returns:
        PayPalSession -- A paypal session for all the api http requests
    """
    return _OAuthSession(mode, token)

def authenticate(client_id: str, secret: str, mode: SessionMode, auth_type: AuthType=AuthType.REFRESHABLE, **kwargs) -> PayPalSession:
    """Creates a session for a given user. If a session handles any kind of 
       flags it can be received as a kwarg. 
       
       Supported flags -> 'refresh_limit' for refreshable sessions.

    Arguments:
        client_id {str} -- paypal client id
        secret {str} -- paypal client secret
        mode {SessionMode} -- Desired session mode (LIVE or SANDBOX)
        auth_type {AuthType} -- Desired session type (BASIC, TOKEN, REFRESHABLE)

    Raises:
        IdentityError: If the user fails to authenticate
    
    Returns:
        PayPalSession -- A paypal session for all the api http requests
    """
    if auth_type == AuthType.TOKEN:
        return session_from_token(_authenticate(client_id, secret, mode), mode)    
    token = _authenticate(client_id, secret, mode)
    if auth_type == AuthType.BASIC:
        return _BasicAuthSession(mode, token, client_id, secret)
    return _RefreshableSession(mode, token, client_id, secret, kwargs.get('refresh_limit'))



"""
    Checkout Order resource clients for the paypal resource group.

    Resource docs & Reference: https://developer.paypal.com/docs/api/orders/v2/  
"""
import json

from datetime import datetime
from typing import Type, TypeVar, List

from pypaypal.clients.base import ClientBase, ActionLink

from pypaypal.http import ( 
    parse_url,
    PayPalSession,
    LIVE_API_BASE_URL,
    SANDBOX_API_BASE_URL
)

from pypaypal.entities.base import ( 
    ResponseType, 
    PaypalApiResponse, 
    PatchUpdateRequest,
    PaypalApiBulkResponse
)

from pypaypal.entities.orders import Order, PaymentSource

"""
    Base Resource Live URL
"""
_LIVE_RESOURCE_BASE_URL = parse_url(LIVE_API_BASE_URL, 'checkout', 'orders')

"""
    Base Resource Sandbox URL
"""
_SANDBOX_RESOURCE_BASE_URL = parse_url(SANDBOX_API_BASE_URL, 'checkout', 'orders')


T = TypeVar('T', bound = 'OrderClient')

class OrderClient(ClientBase):
    """Orders resource group client class.
    """
    def __init__(self, url: str, session: PayPalSession):
        super().__init__(url, session)

    def create_order(self, order: Order, request_id: str = None, partner_attr_id: str = None, response_type: ResponseType = ResponseType.MINIMAL) -> PaypalApiResponse[Order]:
        """Calls the paypal Api to create an order
        
        Arguments:
            order {Order} -- The order information.
        
        Keyword Arguments:
            request_id {str} -- Paypal request id for idempotency (default: {None})
            partner_attr_id {str} --  Identifies the caller as a PayPal partner. To receive revenue attribution. (default: {None})
            response_type {ResponseType} -- Response type to be handled (default: {ResponseType.MINIMAL})
        
        Returns:
            PaypalApiResponse[Order] -- An api response with the order
        """
        url = self._base_url
        
        headers = { 'Prefer': response_type.as_header_value() }

        if request_id:
            headers['PayPal-Request-Id'] = request_id
        if partner_attr_id:
            headers['PayPal-Partner-Attribution-Id'] = partner_attr_id
        
        api_response = self._session.post(url, json.dumps(order.to_dict()), headers = headers)

        if api_response.status_code // 100 != 2:
            return PaypalApiResponse(True, api_response)

        return PaypalApiResponse(False, api_response, Order.serialize_from_json(api_response.json(), response_type))

    def update_order(self, order_id: str, updates: List[PatchUpdateRequest]) -> PaypalApiResponse:
        """Updates an order with the CREATED or APPROVED status.
        
        Arguments:
            order_id {str} -- The id of the order to be updated.
            updates {List[PatchUpdateRequest]} -- The updates to be made.
        
        Returns:
            PaypalApiResponse -- API operation response
        """
        url = parse_url(self._base_url, order_id)
        body = json.dumps([{ 'op': x.operation, 'path': x.path, 'value': x.value } for x in updates ])

        api_response = self._session.patch(url, body)

        if api_response.status_code != 204:
            return PaypalApiResponse(True, api_response)

        return PaypalApiResponse(False, api_response)

    def show_order_details(self, order_id: str) -> PaypalApiResponse[Order]:
        """Calls the api to retrieve the order details
        
        Arguments:
            order_id {str} -- The id of the order.
        
        Returns:
            PaypalApiResponse -- API operation response with the order if successful
        """
        api_response = self._session.get(parse_url(self._base_url, order_id))

        if api_response.status_code != 200:
            return PaypalApiResponse(True, api_response)

        return PaypalApiResponse(False, api_response, Order.serialize_from_json(api_response.json()))

    def authorize_payment_for_order(
        self, order_id: str, payment_source: PaymentSource = None,
        request_id: str = None,  client_metadata_id: str = None,
        auth_assertion: str = None, response_type: ResponseType = ResponseType.MINIMAL) -> PaypalApiResponse[Order]:
        """Calls the API to Authorize a payment for an order. 
           To successfully authorize payment for an order, 
           the buyer must first approve the order 
           or a valid payment_source must be provided in the request.
        
        Arguments:
            order_id {str} -- The order id
        
        Keyword Arguments:
            payment_source {PaymentSource} -- Source of payment for the order if the user wasn't redirected in the order creation to approve the payment. (default: {None})
            request_id {str} -- Request id for idempotence (default: {None})
            client_metadata_id {str} -- Verifies that the payment originates from a valid, user-consented device and application. Must be included in order to be eligible for PayPal Seller Protection. (default: {None})
            auth_assertion {str} -- A JWT assertion that identifies the merchant (default: {None})
            response_type {ResponseType} --  (default: {ResponseType.MINIMAL})
        
        Returns:
            PaypalApiResponse[Order] -- API operation response with the order if successful
        """
        url = parse_url(self._base_url, order_id, 'authorize')
        
        headers = { 'Prefer': response_type.as_header_value() }

        if request_id:
            headers['PayPal-Request-Id'] = request_id
        if  auth_assertion:
            headers['PayPal-Auth-Assertion'] = auth_assertion
        if client_metadata_id:
            headers['PayPal-Client-Metadata-Id'] = client_metadata_id

        if payment_source:
            api_response = self._session.post(url, json.dumps(payment_source.to_dict()), headers = headers)
        else:
            api_response = self._session.post(url, None)

        if api_response.status_code // 100 != 2:
            return PaypalApiResponse(True, api_response)
        
        return PaypalApiResponse(False, api_response, Order.serialize_from_json(api_response.json(), response_type))

    def capture_payment_for_order(
        self, order_id: str, payment_source: PaymentSource = None,
        request_id: str = None,  client_metadata_id: str = None,
        auth_assertion: str = None, response_type: ResponseType = ResponseType.MINIMAL) -> PaypalApiResponse[Order]:
        """Calls the API to Capture a payment for an order. 
           To successfully authorize payment for an order, 
           the buyer must first approve the order 
           or a valid payment_source must be provided in the request.
        
        Arguments:
            order_id {str} -- The order id
        
        Keyword Arguments:
            payment_source {PaymentSource} -- Source of payment for the order if the user wasn't redirected in the order creation to approve the payment. (default: {None})
            request_id {str} -- Request id for idempotence (default: {None})
            client_metadata_id {str} -- Verifies that the payment originates from a valid, user-consented device and application. Must be included in order to be eligible for PayPal Seller Protection. (default: {None})
            auth_assertion {str} -- A JWT assertion that identifies the merchant (default: {None})
            response_type {ResponseType} --  (default: {ResponseType.MINIMAL})
        
        Returns:
            PaypalApiResponse[Order] -- API operation response with the order if successful
        """
        url = parse_url(self._base_url, order_id, 'capture')
        
        headers = { 'Prefer': response_type.as_header_value() }

        if request_id:
            headers['PayPal-Request-Id'] = request_id
        if  auth_assertion:
            headers['PayPal-Auth-Assertion'] = auth_assertion
        if client_metadata_id:
            headers['PayPal-Client-Metadata-Id'] = client_metadata_id

        if payment_source:
            api_response = self._session.post(url, json.dumps(payment_source.to_dict()), headers = headers)
        else:
            api_response = self._session.post(url, None)

        if api_response.status_code // 100 != 2:
            return PaypalApiResponse(True, api_response)
        
        return PaypalApiResponse(False, api_response, Order.serialize_from_json(api_response.json(), response_type))

    @classmethod
    def for_session(cls: T, session: PayPalSession) -> T:
        """Creates a client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of Dispute client with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(base_url, session)



"""
    Module with serialized object representations of paypal responses
"""
import copy
from enum import Enum
from datetime import datetime

from abc import ABC, abstractmethod
from collections.abc import Iterable
from email.mime.base import MIMEBase
from typing import Type, TypeVar, List, Generic, Dict, Iterable

import dateutil.parser

from pypaypal.http import PayPalSession
from pypaypal.errors import PaypalRequestError, PayPalErrorDetail

T = TypeVar('T', bound = 'PayPalEntity')

P = TypeVar('P', bound = 'PaypalPortableAddress')

class ResponseType(Enum):
    MINIMAL = 1
    REPRESENTATION = 2
    UNDEFINED = 3

    def is_minimal(self) -> bool:
        """checks if this response type is minimal
        
        Returns:
            bool -- True if MINIMAL False otherwise
        """
        return self == ResponseType.MINIMAL

    def is_representation(self) -> bool:
        """checks if this response type is representation
        
        Returns:
            bool -- True if REPRESENTATION False otherwise
        """
        return self == ResponseType.REPRESENTATION

    def as_header_value(self) -> str:
        """Gets the header value for this return type. E.g: Prefer: return={type} 
        
        Returns:
            str -- the header value
        """
        return 'return=representation' if self.is_representation() else 'return=minimal'

class RequestMethod(Enum):
    """Enumeration for request methods    
    """
    GET = 1
    POST = 2
    PUT = 3
    DELETE = 4
    PATCH = 5

class DisbursementMode(Enum):
    """Payment Disbursement mode
    """
    INSTANT = 1
    DELAYED = 2

class AppCtxPayeePreference(Enum):
    UNRESTRICTED = 1
    IMMEDIATE_PAYMENT_REQUIRED = 2

class AppCtxShippingPreference(Enum):
    GET_FROM_FILE = 1
    NO_SHIPPING = 2
    SET_PROVIDED_ADDRESS = 3

class AppCtxLandingPage(Enum):
    LOGIN = 1
    BILLING = 2
    NO_PREFERENCE = 3

class PayPalEntity(ABC):
    """
        Base class with common properties for serialized paypal entities
    """
    def __init__(self, json_response: dict= dict(), response_type: ResponseType = ResponseType.MINIMAL):
        self._response_type = response_type
        self._json_response = json_response

    @property
    def json_data(self) -> dict:
        """Getter for this instance private json data
        
        Returns:
            dict -- A deep copy of the instance json attribute
        """
        return copy.deepcopy(self._json_response) if self._json_response else dict()

    def to_dict(self) -> dict:
        d = copy.deepcopy(self.__dict__)
        pp_entity_props = { k : v.to_dict() for k,v in d.items() if isinstance(v, PayPalEntity) }
        
        for k,v in d.items():
            if isinstance(v, Iterable) and not isinstance(v, str):
                d[k] = [ x.to_dict() if isinstance(x, PayPalEntity) else x for x in v ]

        return { k : v for k,v in { **d, **pp_entity_props }.items() if v != None and k not in {'_response_type', '_json_response'} }

    @classmethod
    def _build_args(cls, json_data: dict, entity_types: Dict[str, T] = dict(), array_types: Dict[str, Iterable[T]] = dict()) -> dict:
        """Helper class method to build arguments for serialization on child classes
        
        Arguments:
            json_data {dict} -- paypal json response data
        
        Keyword Arguments:
            entity_types {Dict[str, T]} -- Serializable child class single attributes (default: {dict()})
            array_types {Dict[str, Iterable[T]]} -- Serializable child class array attributes (default: {dict()})
        
        Returns:
            dict -- An array with the arguments for serialization.
        """
        return {
            # Primitives
            **json_data,
            # Serialized types
            **{ k : entity_types[k].serialize_from_json(v) for k,v in json_data.items() if k in entity_types and v },
            # Serilized type arrays
            **{ k : [array_types[k].serialize_from_json(v) for v in json_data[k]] for k in json_data.keys() if k in array_types }
        }

    @classmethod
    def instance_from_dict(cls: Type[T], dictionary: dict) -> T:
        """Factory method meant to create an instance from a dictionary
        
        Arguments:
            cls {Type[T]} -- class
            dictionary {dict} -- dictionary with all the needed args
        
        Returns:
            T -- A child class instance. 
        """
        return cls.serialize_from_json(dictionary, response_type = ResponseType.UNDEFINED)

    @classmethod
    @abstractmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        """Serializes a loaded json (json.loads(str)) to a subclass instance
        
        Arguments:
            cls {Type[T]} -- the class
            json_data {dict} -- the json to serialize
            response_type -- flag with the response type defaults to Minimal
        
        Returns:
            T -- new subsclass instance from json
        """
        pass

class ActionLink(PayPalEntity):
    """Wraping class for entities HATEOAS action links 
    """

    def __init__(self, href: str, rel: str, method: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.rel = rel
        self.href = href
        self._method = method

    @property
    def method(self) -> RequestMethod:
        """Getter for the request method
        
        Returns:
            RequestMethod -- Enumerated constant for the instance request method
        """
        return RequestMethod[self._method]

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        """Serializes a json to a subclass instance
        
        Arguments:
            cls {Type[T]} -- the class
            json_data {dict} -- the json to serialize
            response_type -- flag with the response type defaults to Minimal
        
        Returns:
            T -- new subsclass instance from json
        """
        return cls(json_data['href'], json_data['rel'], json_data['method'], json_data = json_data, response_type = response_type)

class PaypalApiResponse(Generic[T]):
    """Response wrapper for api responses
    """
    def __init__(self, errors: bool, api_response, parsed_response: Type[T]=None):
        self.has_errors = errors
        self._raw_response = api_response
        self.parsed_response = parsed_response
    
    @property
    def error_detail(self) -> PayPalErrorDetail:
        """Returns the error details if there was an error in the response
        
        Returns:
            PayPalErrorDetail -- Error details if exists else None
        """
        data = self._raw_response.json()
        return PayPalErrorDetail.serialize_from_json(data) if self.has_errors and data else None
    
    @classmethod
    def success(cls, api_response, parsed_response: Type[T] = None) -> 'PaypalApiResponse':
        """Factory method for successful requests
        """
        return cls(False, api_response, parsed_response)
    
    @classmethod
    def error(cls, api_response, parsed_response: Type[T] = None) -> 'PaypalApiResponse':
        """Factory method for unsuccessful requests
        """
        return cls(True, api_response, parsed_response)

class PaypalApiBulkResponse(Generic[T]):
    """Response wrapper for api responses
    """
    def __init__(self, errors: bool, api_response, parsed_response: List[Type[T]]=None):
        self.errors = errors
        self._raw_response = api_response
        self.parsed_response = parsed_response
    
    @property
    def error_detail(self) -> PayPalErrorDetail:
        """Returns the error details if there was an error in the response
        
        Returns:
            PayPalErrorDetail -- Error details if exists else None
        """
        data = self._raw_response.json()
        return PayPalErrorDetail.serialize_from_json(data) if self.has_errors and data else None

    @classmethod
    def success(cls, api_response, parsed_response: List[Type[T]] = None) -> 'PaypalApiBulkResponse':
        """Factory method for successful requests
        """
        return cls(False, api_response, parsed_response)
    
    @classmethod
    def error(cls, api_response, parsed_response: List[Type[T]] = None) -> 'PaypalApiBulkResponse':
        """Factory method for unsuccessful requests
        """
        return cls(True, api_response, parsed_response)

class PaypalPage(Generic[T]):
    """Response wrapper for paged responses
    """
    def __init__(self, error: bool, api_response, total_items: int, total_pages: int, elements: List[T], links: List[ActionLink]):
        self.errors = error
        self.links = links
        self.elements = elements or []
        self.total_items = total_items
        self.total_pages = total_pages
        self._raw_response = api_response

    @property
    def element_count(self) -> int:
        """Number of elements. It can be less than the page size
        
        Returns:
            int -- Number of elements
        """
        return len(self.elements)

    @property
    def error_detail(self) -> PayPalErrorDetail:
        """Returns the error details if there was an error in the response
        
        Returns:
            PayPalErrorDetail -- Error details if exists else None
        """
        data = self._raw_response.json()
        return PayPalErrorDetail.serialize_from_json(data) if self.error and data else None

    @property
    def next_page_link(self) -> ActionLink:
        """Retrieves a link to read the next page.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'next', self.links), None)
        
    @property
    def last_page_link(self) -> ActionLink:
        """Retrieves a link to read the last page.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'last', self.links), None)

    @property
    def first_page_link(self) -> ActionLink:
        """Retrieves a link to read the last page.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'first', self.links), None)

    @classmethod
    def success(cls, api_response, total_items: int, total_pages: int, elements: List[T], links: List[ActionLink]) -> 'PaypalPage':
        """Factory method for successful requests
        """
        return cls(False, api_response, total_items, total_pages, elements, links)
    
    @classmethod
    def error(cls, api_response, total_items: int = 0, total_pages: int = 0, elements: List[T] = [], links: List[ActionLink] = []) -> 'PaypalPage':
        """Factory method for unsuccessful requests
        """
        return cls(True, api_response, total_items, total_pages, elements, links)

    @classmethod
    def full_parse_success(cls, api_response, entity_class: Type[T], response_entity_key: str, response_type: ResponseType = ResponseType.MINIMAL) -> 'PaypalPage':
        """Fully parses a successful api response to return a page
        """
        j_data = api_response.json()
        links = [ ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in j_data.get('links', []) ]
        elements = [ entity_class.serialize_from_json(x, response_type) for x in j_data.get(response_entity_key, [])]
        return cls(False, api_response, j_data.get('total_items'), j_data.get('total_pages'), elements, links)

class Money(PayPalEntity):
    """Amount object definition for paypal request/responses
    """

    def __init__(self, currency_code: str, value: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.value = value
        self.currency_code = currency_code

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(json_data['currency_code'], json_data['value'], json_response= json_data, response_type = response_type)

class PaypalMerchant(PayPalEntity):
    """Merchant object representation
    """
    
    def __init__(self, email: str, merchant_id: str, name: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.name = name
        self.email = email
        self.merchant_id = merchant_id

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(
            json_data['email'], json_data['merchant_id'], json_data['name'],
            json_response= json_data, response_type = response_type
        )

class PaypalMessage(PayPalEntity):
    """Message object representation
    """
    
    def __init__(self, posted_by: str = None, time_posted: str = None, content: str = None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.content = content
        self.posted_by = posted_by
        self._time_posted = time_posted
    
    def to_dict(self) -> dict:
        d = super().to_dict()
        if '_time_posted' in d.keys():
            d['time_posted'] = d.pop('_time_posted')
        return d

    @property
    def time_posted(self) -> datetime:
        try:
            return dateutil.parser.parse(self._time_posted) if self._time_posted else None
        except:
            return None

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(**json_data, json_response= json_data, response_type = response_type)

class AddressLine:
    """Address line wrapper
    """
    def __init__(self, address_line_number: int, address_line_info: str):
        self.addr_number = address_line_number
        self.addr_line_info = address_line_info

class AdminArea:
    """Address line wrapper
    """
    def __init__(self, admin_area_number: int, admin_area_info: str):
        self.adm_number = admin_area_number
        self.adm_area_info = admin_area_info

class Street:
    """Street definition
    """
    def __init__(self, street_no: str, street_name: str, street_type: str):
        self.street_no = street_no
        self.street_name = street_name
        self.street_type = street_type

class PaypalAddressDetail:
    """Non-portable additional address details 
    """
    def __init__(self, street: Street, delivery_service: str, building_name: str, sub_building: str):
        self._street = street
        self.sub_building = sub_building
        self.building_name = building_name 
        self.delivery_service = delivery_service
    
    @property
    def street_number(self) -> str:
        return self._street.street_no if self._street else None

    @property
    def street_name(self) -> str:
        return self._street.street_name if self._street else None

    @property
    def street_type(self) -> str:
        return self._street.street_type if self._street else None

class PaypalPortableAddress(PayPalEntity):
    """Paypal portable Addresses object representation
    """
    def __init__(
            self, country_code: str, postal_code: str, address_lines: List[AddressLine], 
            admin_areas: List[AdminArea], details: PaypalAddressDetail, **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.address_details = details
        self.postal_code = postal_code
        self.country_code = country_code
        self._admin_areas = { x.adm_number : x.adm_area_info for x in admin_areas }
        self._address_lines = { x.addr_number : x.addr_line_info for x in address_lines }

    def _adm_area_for_index(self, index: int) -> str:
        return self._admin_areas[index].adm_area_info if index in self._admin_areas.keys() else None

    def _address_line_for_index(self, index: int) -> str:
        return self._address_lines[index].addr_line_info if index in self._address_lines.keys() else None

    @property
    def address_line_1(self) -> str:
        return self._address_line_for_index(1)

    @property
    def address_line_2(self) -> str:
        return self._address_line_for_index(2)

    @property
    def address_line_3(self) -> str:
        return self._address_line_for_index(3)

    @property
    def admin_area_1(self) -> str:
        return self._adm_area_for_index(1)

    @property
    def admin_area_2(self) -> str:
        return self._adm_area_for_index(2)

    @property
    def admin_area_3(self) -> str:
        return self._adm_area_for_index(3)
    
    @property
    def admin_area_4(self) -> str:
        return self._adm_area_for_index(4)

    def to_dict(self) -> dict:
        d = {
            'postal_code': self.postal_code, 'country_code': self.country_code,
            'admin_area_1': self.admin_area_1, 'admin_area_2': self.admin_area_2, 
            'admin_area_3': self.admin_area_3, 'admin_area_4': self.admin_area_4,
            'address_line_1': self.address_line_1, 'address_line_2': self.address_line_2,
            'address_line_3': self.address_line_3
        }
        return { k:v for k,v in d.items() if v != None }
        
    @classmethod
    def serialize_from_json(cls: Type[P], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> P:
        adm_areas, addr_lines, details = [], [], None
        
        for i in range(1, 5):
            adm = 'admin_area_{}'.format(i)
            if adm in json_data.keys():
                adm_areas.append(AdminArea(i, json_data[adm]))
            if i <= 4:
                addr = 'address_line_{}'.format(i)
                if addr in json_data.keys():
                    addr_lines.append(AddressLine(i, json_data[addr]))

        j_details = json_data.get('address_details')

        if j_details:
            street = Street(j_details.get('street_number'), j_details.get('street_name'), j_details.get('street_type'))
            details = PaypalAddressDetail(street, j_details.get('delivery_service'), j_details.get('building_name'), j_details.get('sub_building'))

        return cls(json_data.get('country_code'), json_data.get('postal_code'), addr_lines, adm_areas, details)
    
class PatchUpdateRequest:
    def __init__(self, path: str, value: str, operation: str):
        self.path = path
        self.value = value
        self.operation = operation

class PaypalName(PayPalEntity):
    """Name object representation
    """

    def __init__(self, given_name: str, surname: str,  **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.surname = surname
        self.given_name = given_name
        self.prefix = self._json_response.get('prefix', kwargs.get('prefix'))
        self.suffix = self._json_response.get('suffix', kwargs.get('suffix'))
        self.full_name = self._json_response.get('full_name', kwargs.get('full_name'))
        self.middle_name = self._json_response.get('middle_name', kwargs.get('middle_name'))

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(
            json_data.get('given_name'), json_data.get('surname'), json_response= json_data, response_type = response_type
        )

class PaypalPhoneDetail(PayPalEntity):
    """Phone detail object representation
    """

    def __init__(self, country_code: str, national_number: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.country_code = country_code
        self.national_number = national_number
        self.phone_type = self._json_response.get('phone_type', kwargs.get('phone_type'))
        self.extension_number = self._json_response.get('extension_number', kwargs.get('extension_number'))

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(
            json_data.get('country_code'), json_data.get('national_number'), json_response= json_data, response_type = response_type
        )
    
class ContactInformation(PayPalEntity):
    """Contact information object representation
    """
    def __init__(self, business_name: str, name: PaypalName = None, address: PaypalPortableAddress = None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.name = name
        self.address = address
        self.business_name = business_name

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        address = PaypalPortableAddress.serialize_from_json(json_data)
        name = PaypalName.serialize_from_json(json_data['name']) if 'name' in json_data.keys() else None

        return cls(json_data['business_name'], name, address, json_response= json_data, response_type = response_type)

class BillingInfo(PayPalEntity):
    """Billing info object representation
    """
    def __init__(self, business_name: str, name: PaypalName = None, address: PaypalPortableAddress = None, phones: List[PaypalPhoneDetail] = [], **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.name = name
        self.address = address
        self.phones = phones or []
        self.business_name = business_name
        self.language = self._json_response.get('language', kwargs.get('language'))
        self.email_address = self._json_response.get('email_address', kwargs.get('email_address'))
        self.additional_info = self._json_response.get('additional_info', kwargs.get('additional_info'))

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        phones = []
        address = PaypalPortableAddress.serialize_from_json(json_data)
        name = PaypalName.serialize_from_json(json_data['name']) if 'name' in json_data.keys() else None

        j_phones = json_data.get('phones')

        if j_phones:
            phones = [ PaypalPhoneDetail.serialize_from_json(x, response_type) for x in j_phones ]

        return cls(
            json_data['business_name'], name, address, phones, json_response= json_data, response_type = response_type
        )

class RecipientInfo(PayPalEntity):
    """Recipient info object representation
    """
    def __init__(self, billing_info: BillingInfo, shipping_info: ContactInformation, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.billing_info = billing_info
        self.shipping_info = shipping_info
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        billing = BillingInfo.serialize_from_json(json_data['billing_info']) if 'billing_info' in json_data.keys() else None
        shipping = ContactInformation.serialize_from_json(json_data['shipping_info']) if 'shipping_info' in json_data.keys() else None

        return cls(billing, shipping, json_response= json_data, response_type = response_type)

class Tax(PayPalEntity):
    """Tax object representation
    """

    def __init__(self, name: str, percent: str, amount: Money, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.name = name
        self.amount = amount
        self.percent = percent

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        amount = Money.serialize_from_json(json_data['amount']) if 'amount' in json_data.keys() else None
        return cls(json_data['name'], amount, json_data['pecent'], json_response= json_data, response_type = response_type)

class Discount(PayPalEntity):
    
    def __init__(self, percent: str, amount: Money, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.amount = amount
        self.percent = percent

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        amount = Money.serialize_from_json(json_data['amount']) if 'amount' in json_data.keys() else None
        return cls(json_data['pecent'], amount, json_response= json_data, response_type = response_type)

class Item(PayPalEntity):
    """Item object representation
    """

    def __init__(self, id: str, name: str, quantity: str, unit_amount: str, tax: Tax = None, discount: Discount = None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.id = id 
        self.tax = tax
        self.name = name
        self.quantity = quantity 
        self.discount = discount 
        self.unit_amount = unit_amount
        self._item_date = self._json_response.get('item_date', kwargs.get('item_date'))
        self.description = self._json_response.get('description', kwargs.get('description'))
        self.unit_of_measure = self._json_response.get('unit_of_measure', kwargs.get('unit_of_measure'))

    @property
    def item_date(self) -> datetime:
        try:
            return dateutil.parser.parse(self._item_date) if self._item_date else None
        except:
            return None

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret['item_date'] = ret.pop('_item_date', None)
        return { k:v for k,v in ret if v != None }

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(
                json_data['id'], json_data['name'], json_data['quantity'], json_data['unit_amount'],
                json_response= json_data, response_type = response_type
            )

class RefundDetail(PayPalEntity):
    """Refund detail object representation
    """

    def __init__(self, rf_type: str, refund_id: str, method: str, refund_date: str, amount: Money, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.type = rf_type
        self.refund_id = refund_id
        self.method = method
        self.amount = amount        
        self._refund_date = refund_date

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret['refund_date'] = ret.pop('_refund_date', None)
        return { k:v for k,v in ret if v != None }

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        amount = Money.serialize_from_json(json_data['amount']) if 'amount' in json_data.keys() else None
        
        return cls(
            json_data['type'], json_data['refund_id'], json_data['method'], json_data['refund_date'], amount,
            json_response= json_data, response_type = response_type
        )
    
    @classmethod
    def for_invoice(cls, refund_date: datetime, amount: Money, method: str) -> T:
        return cls(None, None, method, refund_date.strftime('%Y-%m-%d'), amount)
        
class Refund(PayPalEntity):
    """Refund object representation
    """

    def __init__(self, refund_amount: Money, transactions: List[RefundDetail], **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.refund_amount = refund_amount
        self.transactions = transactions

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        refund_amount, transactions = None, list()

        if 'refund_amount' in json_data.keys():
            refund_amount = Money.serialize_from_json(json_data['refund_amount'], response_type)        
        if 'transactions' in json_data.keys():
            transactions = [RefundDetail.serialize_from_json(x, response_type) for x in json_data['transactions']]
        
        return cls(refund_amount, transactions, json_response= json_data, response_type = response_type)

class AmountRange:
    """Amount range for requests
    """
    def __init__(self, lower_amount: Money, upper_amount: Money):
        self.lower_amount = lower_amount
        self.upper_amount = upper_amount
    
    def to_dict(self) -> dict:
        return { 
            'lower_amount': self.lower_amount.to_dict(), 
            'upper_amount': self.upper_amount.to_dict() 
        }

class DateRange:
    """Date range for request
    """
    def __init__(self, start: str, end: str):
        self.start = start
        self.end = end
    
    def to_dict(self) -> dict:
        return { 'start': self.start, 'end': self.end }

class PayeeBase(PayPalEntity):
    """Payee base object representation.
    """
    def __init__(self, email: str, merchant_id: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.email_address = email
        self.merchant_id = merchant_id

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(
            json_data['email_address'], json_data['merchant_id'],
            json_response= json_data, response_type = response_type
        )

class Payee(PayeeBase):
    """Payee object representation.
    """
    def __init__(self, email: str, merchant_id: str, **kwargs):
        super().__init__(email, merchant_id, **kwargs)

class PlatformFee(PayPalEntity):
    """Platform Fee object representation.
    """
    def __init__(self, amount: Money, payee: PayeeBase, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.payee = payee
        self.amount = amount
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        amount, payee = None, None

        if 'amount' in json_data.keys():
            amount = Money.serialize_from_json(json_data['amount'], response_type)
        if 'payee' in json_data.keys():
            payee = PayeeBase.serialize_from_json(json_data['payee'], response_type)

        return cls(amount, payee, json_response= json_data, response_type = response_type)

    @classmethod
    def create(cls, amount: Money, payee: PayeeBase = None) -> 'PlatformFee':
        return cls(amount, payee)

class PaymentInstruction(PayPalEntity):
    """Payment instruction object representation.
    """

    def __init__(self, platform_fees: List[PlatformFee], disbursement_mode: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.platform_fees = platform_fees
        self.disbursement_mode = disbursement_mode

    @property
    def disbursement_mode_enum(self) -> DisbursementMode:
        try:
            return DisbursementMode[self.disbursement_mode] if self.disbursement_mode else None
        except:
            return None

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        platform_fees = []

        if 'platform_fees' in json_data.keys():
            platform_fees = [PlatformFee.serialize_from_json(x, response_type) for x in json_data['platform_fees']]

        return cls(platform_fees, json_data['disbursement_mode'], json_response= json_data, response_type = response_type)

    @classmethod
    def create(cls, platform_fees: List[PlatformFee], disbursement_mode: DisbursementMode = DisbursementMode.INSTANT):
        return cls(platform_fees, disbursement_mode.name)

class PaymentMethod(PayPalEntity):
    """App context payment method obj representation
    """

    def __init__(self, payer_selected: str, payee_preferred: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.payer_selected = payer_selected
        self.payee_preferred = payee_preferred

    @property
    def payee_preferred_enum(self) -> AppCtxPayeePreference:
        try:
            return AppCtxPayeePreference[self.payee_preferred] if self.payee_preferred else None
        except:
            return None

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(json_data['payer_selected'], json_data['payee_preferred'], json_response= json_data, response_type = response_type)
    
    @classmethod
    def create(cls,  payee_preferred: AppCtxPayeePreference, payer_selected: str = 'PAYPAL') -> 'PaymentMethod':
        return cls(payer_selected, payee_preferred.name)

class ApplicationContext(PayPalEntity):
    """Paypal application context object representation
    """

    _ENTITY_TYPES = { 'payment_method': PaymentMethod }

    def __init__(
        self, brand_name: str = None, locale: str = None, 
        landing_page: str = None, shipping_preference: str = None, 
        user_action: str = None, payment_method: PaymentMethod = None, 
        return_url: str = None, cancel_url: str = None, **kwargs
    ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.brand_name = brand_name
        self.locale = locale
        self.landing_page = landing_page
        self.shipping_preference = shipping_preference
        self.user_action = user_action
        self.payment_method = payment_method
        self.return_url  = return_url
        self.cancel_url = cancel_url
    
    @property
    def landing_page_enum(self) -> AppCtxLandingPage:
        try:
            return AppCtxLandingPage[self.landing_page] if self.landing_page else None
        except:
            return None

    @property
    def shipping_preference_enum(self) -> AppCtxShippingPreference:
        try:
            return AppCtxShippingPreference[self.shipping_preference] if self.shipping_preference else None
        except:
            return None

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = super()._build_args(json_data, cls._ENTITY_TYPES)
        return cls(**args, json_response= json_data, response_type = response_type)
    
    @classmethod
    def create(cls, brand_name: str = None, locale: str = None, return_url: str = None, cancel_url: str = None, 
        payment_method: PaymentMethod = None, landing_page: AppCtxLandingPage = AppCtxLandingPage.NO_PREFERENCE,
        shipping_preference: AppCtxShippingPreference = AppCtxShippingPreference.GET_FROM_FILE, 
        user_action: str = None) -> 'ApplicationContext':
        return cls(
            brand_name, locale, landing_page.name, shipping_preference.name, 
            user_action, payment_method, return_url, cancel_url
        )


from enum import Enum
from datetime import datetime
from typing import Type, List

import dateutil.parser

from pypaypal.entities.base import (
    T, 
    Money, 
    PaypalName,
    ActionLink, 
    PayPalEntity, 
    ResponseType,
    PaymentMethod,
    AppCtxLandingPage,
    ApplicationContext,
    PaypalPortableAddress,
    AppCtxShippingPreference
)

from pypaypal.entities.subscriptions.common import BillingCycleTenureType

class SubscriptionTransactionStatus(Enum):
    # The funds for this captured payment were credited to the payee's PayPal account.
    COMPLETED = 1
    # The funds could not be captured.
    DECLINED = 2
    # An amount less than this captured payment's amount was partially refunded to the payer.
    PARTIALLY_REFUNDED = 3
    # The funds for this captured payment was not yet credited to the payee's PayPal account.
    PENDING = 4
    # An amount greater than or equal to this captured payment's amount was refunded to the payer.
    REFUNDED = 5

class SubscriptionStatus(Enum):
    # The subscription is created but not yet approved by the buyer.
    APPROVAL_PENDING = 1
    # The buyer has approved the subscription.
    APPROVED = 2
    # The subscription is active.
    ACTIVE = 3
    # The subscription is suspended.
    SUSPENDED = 4
    # The subscription is cancelled.
    CANCELLED = 5
    # The subscription is expired.
    EXPIRED = 6

class SubscriptionUserAction(Enum):
    # After you redirect the customer to the PayPal subscription consent page, a Continue button appears.
    # Use this option when you want to control the activation of the subscription and do not want PayPal to activate the subscription.
    CONTINUE = 1
    # After you redirect the customer to the PayPal subscription consent page, a Subscribe Now button appears. 
    # Use this option when you want PayPal to activate the subscription.
    SUBSCRIBE_NOW = 2

class Email(PayPalEntity):
    """Subscription email obj representation
    """
    def __init__(self, email: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.email = email

    @classmethod
    def create(cls, email: str) -> 'Email':
        return cls(email)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(**json_data, json_response= json_data, response_type = response_type)

class SubscriptionApplicationContext(ApplicationContext):
    """Paypal subscription app context object representation
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    
    @property
    def subscription_user_action(self) -> SubscriptionUserAction:
        try:
            return SubscriptionUserAction[self.user_action] if self.user_action else None
        except:
            return None

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return super().serialize_from_json(json_data, response_type)
            
    @classmethod
    def create(cls, brand_name: str = None, locale: str = None, return_url: str = None, cancel_url: str = None, 
        payment_method: PaymentMethod = None, landing_page: AppCtxLandingPage = AppCtxLandingPage.NO_PREFERENCE,
        shipping_preference: AppCtxShippingPreference = AppCtxShippingPreference.GET_FROM_FILE, 
        user_action: SubscriptionUserAction = SubscriptionUserAction.SUBSCRIBE_NOW) -> 'ApplicationContext':
        return super().create(
            brand_name, locale, landing_page.name, shipping_preference.name,
            user_action.name, payment_method, return_url, cancel_url
        )

class ShippingDetail(PayPalEntity):
    """Shipping detail obj representation
    """
    _ENTITY_TYPES =  {'address': PaypalPortableAddress}

    def __init__(self, name: str = None, address: PaypalPortableAddress = None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.name = name
        self.address = address
    
    @classmethod
    def create(cls, name: str, address: PaypalPortableAddress) -> 'ShippingDetail':
        return cls(name, address)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = super()._build_args(json_data, cls._ENTITY_TYPES)
        return cls(**args, json_response= json_data, response_type = response_type)

class Subscriber(PayPalEntity):
    """Subscriber obj representation
    """
    _ENTITY_TYPES =  { 'email_address': Email, 'shipping_address': ShippingDetail }

    def __init__(self, name: str = None, email_address: Email = None, shipping_address: ShippingDetail = None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.name = name
        self.email_address = email_address
        self.shipping_address = shipping_address

    @classmethod
    def create(cls, name: str = None, email_address: Email = None, shipping_address: ShippingDetail = None) -> 'Subscriber':
        return cls(name, email_address, shipping_address)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = super()._build_args(json_data, cls._ENTITY_TYPES)
        return cls(**args, json_response= json_data, response_type = response_type)

class CycleExecution(PayPalEntity):
    """Subscription cycle execution obj representation
    """

    def __init__(
            self, tenure_type: str = None, sequence: int = None, cycles_completed: int = None, 
            cycles_remaining: int = None, current_pricing_scheme_version: int = None,
            total_cycles: int = None, **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.sequence = sequence
        self.tenure_type = tenure_type
        self.total_cycles = total_cycles
        self.cycles_completed = cycles_completed 
        self.cycles_remaining = cycles_remaining
        self.current_pricing_scheme_version = current_pricing_scheme_version
    

    @property
    def tenure_type_enum(self) -> BillingCycleTenureType:
        try:
            return BillingCycleTenureType[self.tenure_type] if self.tenure_type else None
        except:
            return None
    
    @classmethod
    def create(
            cls, *, tenure_type: BillingCycleTenureType, sequence: int, 
            cycles_completed: int, cycles_remaining: int = None, 
            current_pricing_scheme_version: int = None, total_cycles: int = None
        ) -> 'CycleExecution':
        return cls(
            tenure_type.name, sequence, cycles_completed, cycles_remaining, 
            current_pricing_scheme_version, total_cycles
        )

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:        
        return cls(**json_data, json_response= json_data, response_type = response_type)

class LastPaymentDetails(PayPalEntity):
    """Subscription last payment details obj representation.
    """
    
    _ENTITY_TYPES = { 'amount': Money }

    def __init__(self, amount: Money, time: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.time = time
        self.amount = amount

    @property
    def time_as_date(self) -> datetime:
        try:
            return dateutil.parser.parse(self.time) if self.time else None
        except:
            return None

    @classmethod
    def create(cls, *, amount: Money, time: str) -> 'LastPaymentDetails':
        return cls(amount, time)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:        
        args = super()._build_args(json_data, cls._ENTITY_TYPES)
        return cls(**args, json_response= json_data, response_type = response_type)

class BillingInfo(PayPalEntity):
    """Subscription billing info obj representation
    """
    
    # Serializable paypal entity arrays
    _ARRAY_TYPES = { 'cycle_executions': CycleExecution }

    # Serializable simple paypal entities
    _ENTITY_TYPES = { 'outstanding_balance': Money, 'last_payment': LastPaymentDetails}
    
    def __init__(self, failed_payments_count: int, outstanding_balance: Money, last_payment: LastPaymentDetails = None, cycle_executions: List[CycleExecution] = [], **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.last_payment = last_payment
        self.cycle_executions = cycle_executions or []
        self.outstanding_balance = outstanding_balance
        self.failed_payments_count = failed_payments_count

    @classmethod
    def create(cls, *, failed_payments_count: int, outstanding_balance: Money, last_payment: LastPaymentDetails = None, cycle_executions: List[CycleExecution] = []) -> 'BillingInfo':
        return cls(failed_payments_count, outstanding_balance, last_payment, cycle_executions or [])

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:        
        args = super()._build_args(json_data, cls._ENTITY_TYPES, cls._ARRAY_TYPES)
        return cls(**args, json_response= json_data, response_type = response_type)

class Subscription(PayPalEntity):
    """Paypal subscription obj representation
    """
    
    _ENTITY_TYPES = { 'subscriber': Subscriber, 'billing_info': BillingInfo, 'shipping_amount': Money }

    def __init__(
            self, subscription_id: str = None, status: str = None, 
            plan_id: str = None, status_change_note: str = None, 
            quantity: str = None, subscriber: Subscriber = None,             
            billing_info: BillingInfo = None, shipping_amount: Money = None,            
            application_context: SubscriptionApplicationContext = None, **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.id = subscription_id
        self.status = status
        self.plan_id = plan_id 
        self.quantity = quantity
        self.subscriber = subscriber 
        self.billing_info = billing_info
        self.shipping_amount = shipping_amount 
        self.status_change_note = status_change_note
        self.application_context = kwargs.get('application_context')
        self._update_time = self._json_response.get('update_time', kwargs.get('update_time'))
        self._create_time = self._json_response.get('create_time', kwargs.get('create_time'))
        self._start_time = self._json_response.get('_start_time', kwargs.get('_start_time'))            
        self._status_update_time = self._json_response.get('_status_update_time', kwargs.get('_status_update_time'))
        self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]

    @property
    def update_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._update_time) if self._update_time else None
        except:
            return None

    @property
    def create_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._create_time) if self._create_time else None
        except:
            return None

    @property
    def start_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._start_time) if self._start_time else None
        except:
            return None

    @property
    def status_update_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._status_update_time) if self._status_update_time else None
        except:
            return None

    @property
    def status_enum(self) -> SubscriptionStatus:
        """Status of the subscription as an enum constant
        
        Returns:
            PlanStatus -- An enumerated constant representing the subscription status or None
        """
        try:
            return SubscriptionStatus[self.status] if self.status else None
        except:
            return None

    @property
    def read_link(self) -> ActionLink:
        """Retrieves a link to read this entity details.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'self', self.links), None)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        if '_start_time' in ret.keys():
            ret['start_time'] = ret.pop('_start_time')
        return ret

    @classmethod
    def create(
            cls, *, plan_id: str, start_time: datetime = datetime.now(), 
            quantity: str = None, shipping_amount: Money = None,
            subscriber: Subscriber = None, application_context: SubscriptionApplicationContext = None
        ) -> 'Subscription':
        return cls(
            plan_id = plan_id, start_time = start_time.strftime('%Y-%m-%dT%H:%M:%S'), quantity = quantity,
            shipping_amount = shipping_amount, subscriber = subscriber, 
            application_context = application_context
        )

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:        
        args = super()._build_args(json_data, cls._ENTITY_TYPES)
        args['subscription_id'] = args.pop('id', None)
        return cls(**args, json_response= json_data, response_type = response_type)


class AmountWithBreakdown(PayPalEntity):
    """Subscription transaction amount with breakdown
    """
    
    _ENTITY_TYPES = { 
        'net_amount': Money, 'gross_amount': Money, 'fee_amount': Money, 
        'shipping_amount': Money, 'tax_amount': Money
    }
    
    def __init__(
            self, net_amount: Money, gross_amount: Money, 
            fee_amount: Money = None, shipping_amount: Money = None, 
            tax_amount: Money = None, **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.net_amount = net_amount
        self.fee_amount = fee_amount
        self.tax_amount = tax_amount
        self.gross_amount = gross_amount
        self.shipping_amount = shipping_amount

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:        
        args = super()._build_args(json_data, cls._ENTITY_TYPES)
        return cls(**args, json_response= json_data, response_type = response_type)

class SubscriptionTransaction(PayPalEntity):
    """A subscription transaction for querying purposes
    """

    _ENTITY_TYPES = { 
        'payer_name': PaypalName, 'amount_with_breakdown': AmountWithBreakdown
    }

    def __init__(
            self, status: str = None, transaction_id: str = None, 
            amount_with_breakdown: AmountWithBreakdown = None, 
            payer_name: PaypalName = None, payer_email: str = None, 
            **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.status = status 
        self.payer_name = payer_name
        self.payer_email = payer_email
        self.transaction_id = transaction_id
        self.amount_with_breakdown = amount_with_breakdown
        self._time = self._json_response.get('time', kwargs.get('time'))

    @property
    def status_enum(self) -> SubscriptionTransactionStatus:
        """Status of the subscription transaction as an enum constant
        
        Returns:
            PlanStatus -- An enumerated constant representing the subscription status or None
        """
        try:
            return SubscriptionTransactionStatus[self.status] if self.status else None
        except:
            return None

    @property
    def time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._time) if self._time else None
        except:
            return None

    def to_dict(self) -> dict:
        r = super().to_dict()
        r['time'] = r.pop('_time', None)
        r['id'] = r.pop('transaction_id', None)
        return r

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:        
        args = super()._build_args(json_data, cls._ENTITY_TYPES)
        args['transaction_id'] = args.pop('id', None)
        return cls(**args, json_response= json_data, response_type = response_type)



"""
    Payment authorizations resource related entities
"""

from enum import Enum
from datetime import datetime
from typing import Type, List

import dateutil.parser

from pypaypal.entities.base import (
    T,
    Money,
    ActionLink,
    PlatformFee,
    ResponseType,    
    PayPalEntity
)

from pypaypal.entities.payments.commons import SellerProtection

class StatusDetailReason(Enum):
    PENDING_REVIEW = 1 # Authorization is pending manual review

class AuthorizationStatus(Enum):
    CREATED = 1
    CAPTURED = 2
    DENIED = 3
    EXPIRED = 4
    PARTIALLY_CAPTURED = 5
    VOIDED = 6
    PENDING = 7

class AuthStatusDetail(PayPalEntity):
    """Authorization status details
    """
    def __init__(self, reason: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.reason  = reason

    @property
    def reason_enum(self) -> StatusDetailReason:
        try:
            return StatusDetailReason[self.reason] if self.reason else None
        except:
            return None

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(json_data.get('reason'), json_response = json_data, response_type = response_type)

    @classmethod
    def create(cls, reason: StatusDetailReason = StatusDetailReason.PENDING_REVIEW) -> 'AuthStatusDetail':
        return cls(reason.name)

class Authorization(PayPalEntity):
    """Payment authorizatio obj representation
    """

    def __init__(
        self, auth_id: str, status: str, amount: Money, status_details: AuthStatusDetail,
        invoice_id: str, custom_id: str, seller_protection: SellerProtection, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.id = auth_id
        self.status = status
        self.amount = amount
        self.status_details = status_details
        self.invoice_id = invoice_id
        self.custom_id = custom_id
        self.seller_protection = seller_protection
        self._create_time = self._json_response.get('create_time', kwargs.get('create_time'))
        self._update_time = self._json_response.get('update_time', kwargs.get('update_time'))
        self._expiration_time = self._json_response.get('expiration_time', kwargs.get('expiration_time'))
        self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]
   
    @property
    def create_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._create_time) if self._create_time else None
        except:
            return None

    @property
    def update_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._update_time) if self._update_time else None
        except:
            return None

    @property
    def expiration_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._expiration_time) if self._expiration_time else None
        except:
            return None

    @property
    def Status_enum(self) -> AuthorizationStatus:
        try:
            return AuthorizationStatus[self.status] if self.status else None
        except:
            return None

    @property
    def read_link(self) -> ActionLink:
        """Retrieves a link to read this entity details.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'self', self.links), None)

    @property
    def capture_link(self) -> ActionLink:
        """Retrieves a link to perform a capture on this authorized payment.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'capture', self.links), None)

    @property
    def void_link(self) -> ActionLink:
        """Retrieves a link to void this authorization
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'void', self.links), None)

    @property
    def reauthorize_link(self) -> ActionLink:
        """Retrieves a link to reauthorize this payment.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'reauthorize', self.links), None)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        amount, seller_protection, status_details = None, None, None
    
        if 'amount' in json_data.keys():
            amount = Money.serialize_from_json(json_data['amount'], response_type)
        if 'seller_protection' in json_data.keys():
            seller_protection = SellerProtection.serialize_from_json(json_data['seller_protection'], response_type)
        if 'status_details' in json_data.keys():
            status_details = AuthStatusDetail.serialize_from_json(json_data['status_details'], response_type)

        return cls(
            json_data.get('id'), json_data.get('status'), 
            amount, status_details, json_data.get('invoice_id'), json_data.get('custom_id'),
            seller_protection, json_response = json_data, response_type = response_type
        )



"""
    Subscription resource client for the paypal resource group.

    Resource docs & Reference: https://developer.paypal.com/docs/api/subscriptions/v1/
"""
import json

from datetime import datetime
from typing import Type, TypeVar, List, NamedTuple

from pypaypal.entities.base import Money
from pypaypal.clients.base import ClientBase, ActionLink

from pypaypal.http import ( 
    parse_url,
    PayPalSession,
    LEGACY_LIVE_API_BASE_URL,
    LEGACY_SANDBOX_API_BASE_URL
)

from pypaypal.entities.base import ( 
    PaypalPage,
    ResponseType, 
    PaypalApiResponse, 
    PatchUpdateRequest,
    PaypalApiBulkResponse
)

from pypaypal.entities.payments.captures import Capture

from pypaypal.entities.subscriptions.subscriptions import ( 
    Subscription, 
    ShippingDetail,
    SubscriptionTransaction,
    SubscriptionApplicationContext
)

"""
    Base Resource Live URL
"""
_LIVE_RESOURCE_BASE_URL = parse_url(LEGACY_LIVE_API_BASE_URL, 'billing', 'subscriptions')

"""
    Base Resource Sandbox URL
"""
_SANDBOX_RESOURCE_BASE_URL = parse_url(LEGACY_SANDBOX_API_BASE_URL, 'billing', 'subscriptions')

T = TypeVar('T', bound = 'SubscriptionClient')

class SubscriptionQuantityUpdateRequest(NamedTuple):
    """Immutable class to request a subscription quantity update for a product/service.
    """
    plan_id: str
    quantity: str
    shipping_amount: Money
    shipping_address: ShippingDetail
    application_context: SubscriptionApplicationContext

    def to_dict(self) -> dict:
        return {
            'plan_id': self.plan_id,
            'quantity': self.quantity,
            'shipping_amount': self.shipping_amount.to_dict(),
            'shipping_address': self.shipping_address.to_dict(),
            'application_context': self.application_context.to_dict()
        }

class SubscriptionClient(ClientBase):
    """Subscriptions resource group client class.
    """
    def __init__(self, url: str, session: PayPalSession):
        super().__init__(url, session)

    def create_subscription(self, subscription: Subscription, request_id: str = None, response_type: ResponseType = ResponseType.MINIMAL) -> PaypalApiResponse[Subscription]:
        """Calls the paypal API to create a subscription
        
        Arguments:
            subscription {Subscription} -- The subscription to be made (use the create factory method)
        
        Keyword Arguments:
            request_id {str} -- Paypal request id for idempotence (default: {None})
            response_type {ResponseType} -- Peferred response type (default: {ResponseType.MINIMAL})
        
        Returns:
            PaypalApiResponse[Subscription] -- an API response object with the subscription info
        """
        url = self._base_url
        body = json.dumps(subscription.to_dict())
        headers = { 'Prefer': response_type.as_header_value() }

        if request_id:
            headers['PayPal-Request-Id'] = request_id
        
        api_response = self._session.post(url, body, headers = headers)

        if api_response.status_code // 100 != 2:
            return PaypalApiResponse.error(api_response)
        
        return PaypalApiResponse.success(api_response, Subscription.serialize_from_json(api_response.json(), response_type))

    def update_subscription(self, subscription_id: str, update_request: List[PatchUpdateRequest]) -> PaypalApiResponse:
        """Patch request to update a subscription. See the docs for details 
        
        Arguments:
            subscription_id {str} -- id of the subscription to be updated.
            update_request {PatchUpdateRequest} -- request object with update info.
        
        Returns:
            PaypalApiResponse -- Response obj with operation status
        """
        url = parse_url(self._base_url, subscription_id) 
        body = json.dumps([ {'op': x.operation, 'value': x.value, 'path': x.path } for x in update_request ])

        api_response = self._session.patch(url, body)

        if api_response.status_code // 100 != 2:
            return PaypalApiResponse.error(api_response)
        
        return PaypalApiResponse.success(api_response)

    def show_subscription_details(self, subscription_id: str) -> PaypalApiResponse[Subscription]:
        """Calls the paypal API to get a subscription details
        
        Arguments:
            subscription_id {str} -- The subscription identifier
        
        Returns:
            PaypalApiResponse[Subscription] -- Api response obj with the subscription info.
        """
        api_response = self._session.get(parse_url(self._base_url, subscription_id))

        if api_response.status_code // 100 != 2:
            return PaypalApiResponse.error(api_response)
        
        return PaypalApiResponse.success(api_response, Subscription.serialize_from_json(api_response.json()))

    def activate_subscription(self, subscription_id: str) -> PaypalApiResponse:
        """Calls the API to activate a subscription
        
        Arguments:
            subscription_id {str} -- The id of the subscription to activate
        
        Returns:
            PaypalApiResponse -- Response with the operation status
        """
        return self._execute_subscription_action(subscription_id, 'activate')

    def reactivate_subscription(self, subscription_id: str, reason: str) -> PaypalApiResponse:
        """Calls the API to re-activate a subscription
        
        Arguments:
            subscription_id {str} -- The id of the subscription to activate
        
        Returns:
            PaypalApiResponse -- Response with the operation status
        """
        return self._execute_subscription_action(subscription_id, 'activate', reason)
    
    def cancel_subscription(self, subscription_id: str, reason: str) -> PaypalApiResponse:
        """Calls the API to cancel a subscription
        
        Arguments:
            subscription_id {str} -- The id of the subscription to cancel
        
        Returns:
            PaypalApiResponse -- Response with the operation status
        """
        return self._execute_subscription_action(subscription_id, 'cancel', reason)

    def capture_authorized_payment_on_subscription(
            self, subscription_id: str, note: str, amount: Money, 
            capture_type: str = 'OUTSTANDING_BALANCE', request_id: str = None
        ) -> PaypalApiResponse[Capture]:
        """ Calls the Paypal API to perform a capture on an authorized payment 
            from the subscriber on the subscription.
        
        Arguments:
            subscription_id {str} -- the subscription id
            note {str} -- The reason or note for the subscription charge.
            amount {Money} -- The amount of the outstanding balance. Must not be greater than the current outstanding balance amount.
        
        Returns:
            PaypalApiResponse[Capture] -- A response with a capture
        """
        url = parse_url(self._base_url, subscription_id, 'capture')
        
        body = json.dumps({
            'note': note, 
            'amount': amount.to_dict(), 
            'capture_type': capture_type
        })

        if not request_id:
            api_response = self._session.post(url, body)
        else:
            api_response = self._session.post(url, body, headers = { 'PayPal-Request-Id': request_id })
             
        if api_response.status_code // 100 != 2:
            return PaypalApiResponse.error(api_response)
        
        j_data = api_response.json()
        return PaypalApiResponse.success(api_response, Capture.serialize_from_json(j_data) if j_data else None)

    def update_quantity_in_subscription(self, subscription_id: str, update: SubscriptionQuantityUpdateRequest) -> PaypalApiResponse[Subscription]:
        """Calls the paypal API to updates the quantity of the product or service in a subscription. 
           this method can also be used to to switch the plan and update the shipping_amount, 
           shipping_address values for the subscription. This type of update requires the buyer's consent.
        
        Arguments:
            subscription_id {str} -- the subscription id
            update {SubscriptionQuantityUpdateRequest} -- update request info.
        
        Returns:
            PaypalApiResponse[Subscription] -- Response status with subscription details
        """
        body = json.dumps(update.to_dict())
        url = parse_url(self._base_url, subscription_id, 'revise')

        api_response = self._session.post(url, body)
             
        if api_response.status_code // 100 != 2:
            return PaypalApiResponse.error(api_response)
        
        return PaypalApiResponse.success(api_response, Subscription.serialize_from_json(api_response.json()))

    def suspend_subscription(self, subscription_id: str, reason: str) -> PaypalApiResponse:
        """Calls the API to suspend a subscription
        
        Arguments:
            subscription_id {str} -- The id of the subscription to cancel
        
        Returns:
            PaypalApiResponse -- Response with the operation status
        """
        return self._execute_subscription_action(subscription_id, 'suspend', reason)

    def _execute_subscription_action(self, subscription_id: str, action_name: str, reason: str = None) -> PaypalApiResponse:
        """Executes a generic and simple subscription action call to the Paypal API
        
        Arguments:
            subscription_id {str} -- The subscription id
            action_name {str} -- API URL action name
        
        Keyword Arguments:
            reason {str} -- A comment or reason if needed (default: {None})
        
        Returns:
            PaypalApiResponse -- [description]
        """
        body = json.dumps({ 'reason': reason }) if reason else None
        url = parse_url(self._base_url, subscription_id, action_name)

        api_response = self._session.post(url, body)

        if api_response.status_code // 100 != 2:
            return PaypalApiResponse.error(api_response)
        
        return PaypalApiResponse.success(api_response)

    def list_subscription_transactions(self, subscription_id: str, start_time: datetime, end_time: datetime) -> PaypalPage[SubscriptionTransaction]:
        """Calls the API to lists transactions for a subscription.
        
        Arguments:
            subscription_id {str} -- The subscription id
            start_time {datetime} -- transaction start time
            end_time {datetime} -- transaction end time
        
        Returns:
            PaypalPage[SubscriptionTransaction] -- Paged transaction info
        """
        fmt = '%Y-%m-%dT%H:%M:%S'
        url = parse_url(self._base_url, subscription_id, 'transactions')        
        params = { end_time.strftime(fmt), start_time.strftime(fmt) }

        api_response = self._session.get(url, params)

        if api_response.status_code // 100 != 2:
            return PaypalPage.error(api_response)
        
        return PaypalPage.full_parse_success(api_response, SubscriptionTransaction, 'transaction')

    @classmethod
    def for_session(cls: T, session: PayPalSession) -> T:
        """Creates a client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of Dispute client with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(base_url, session)



"""Common entities in payment resources
"""

from enum import Enum
from typing import Type, List

from pypaypal.entities.base import (
    T,
    ResponseType,    
    PayPalEntity
)

class DisputeCategoryReason(Enum):
    ITEM_NOT_RECEIVED = 1
    UNAUTHORIZED_TRANSACTION = 2

class SellerProtectionStatus(Enum):
    ELIGIBLE = 1 # Seller PayPal balance remains intact if a customer claims not receiving an item or an account holder claims not authorizing the payment.
    PARTIALLY_ELIGIBLE = 2 # PayPal balance remains intact if a customer claims not receiving an item.
    NOT_ELIGIBLE = 3 # transaction not eligible for seller protection.

class DisputeCategory(PayPalEntity):
    """Dispute Category obj representation
    """
    def __init__(self, dispute_category: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.dispute_category = dispute_category

    @property
    def dispute_category_reason(self) -> DisputeCategoryReason:
        try:
            return DisputeCategoryReason[self.dispute_category] if self.dispute_category else None
        except:
            return None

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(json_data, json_response = json_data, response_type = response_type)

    @classmethod
    def create(cls, reason: DisputeCategoryReason) -> 'DisputeCategory':
        return cls(reason.name)

class SellerProtection(PayPalEntity):
    """Authorization seller protection details
    """
    def __init__(self, status: str, dispute_categories: List[DisputeCategory],**kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.status = status
        self.dispute_categories = dispute_categories

    @property
    def Status_enum(self) -> SellerProtectionStatus:
        try:
            return SellerProtectionStatus[self.status] if self.status else None
        except:
            return None

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        dispute_categories = []
        
        if 'dispute_categories' in json_data.keys():
            dispute_categories = [DisputeCategory.serialize_from_json(x, response_type) for x in json_data['dispute_categories']]

        return cls(json_data.get('status'), dispute_categories, json_response = json_data, response_type = response_type)

    @classmethod
    def create(cls, status: SellerProtectionStatus, dispute_categories: List[DisputeCategory]) -> 'SellerProtection':
        return cls(status.name, dispute_categories)

class ExchangeRate(PayPalEntity):
    """Capture exchange rate
    """
    def __init__(self, source_currency: str, target_currency: str, value: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.value = value
        self.source_currency = source_currency
        self.target_currency = target_currency

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:        
        return cls(
            json_data.get('source_currency'), json_data.get('target_currency'), 
            json_data.get('value'), json_response = json_data, response_type = response_type
        )



"""
    Module containing the client errors and exceptions
"""

import copy

from typing import Type, TypeVar, List

T = TypeVar('T', bound = 'PayPalErrorDetail')

class PayPalErrorDetail:
    """
        Generic paypal error
    """
    def __init__(self, json_response: dict, name: str, message: str, info_link: str):
        self.name = name
        self.message = message
        self.information_link = info_link
        self._json_response = json_response

    @property
    def as_json_dict(self) -> dict:
        return self._json_response

    @property
    def debug_id(self) -> str:
        return self._json_response.get('debug_id')

    @property
    def additional_details(self) -> List[dict]:
        return self._json_response.get('details')

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict) -> T:
        return cls(json_data, json_data.get('name'), json_data.get('message'), json_data.get('information_link'))

class IdentityError(Exception):
    """
        Authentication errors
    """
    def __init__(self, response):
        super().__init__('Identity error occoured')
        self.response = response
        self.error = response.json().get('error')
        self.error_description = response.json().get('error_description')

class ExpiredSessionError(Exception):
    """
        Errors regarding expired tokens & session
    """
    def __init__(self, data):
        super().__init__('token or session has expired')
        self.data = data

class PaypalRequestError(Exception):
    """
        Non identity errors that may arise from a request to the paypal api
    """
    def __init__(self, details: PayPalErrorDetail):
        super().__init__(details.message)
        self.details = details

# class EntityRefreshError(Exception):
#     """
#       Error raised when there's a failure refreshing an entity  
#     """
#     def __init__(self, entity):
#         super().__init__(f'There was an error refreshing an entity of {type(entity)}')




"""
    Module with subscription plans related entities
"""

from enum import Enum
from datetime import datetime
from typing import Type, List

import dateutil.parser

from pypaypal.entities.base import (
    T, 
    Money, 
    ActionLink, 
    PayPalEntity, 
    ResponseType, 
)

from pypaypal.entities.subscriptions.common import BillingCycleTenureType

class PlanStatus(Enum):
    # The plan was created. You cannot create subscriptions for a plan in this state.
    CREATED = 1
    # The plan is inactive.
    INACTIVE = 2
    # The plan is active. You can only create subscriptions for a plan in this state.
    ACTIVE = 3

class FrequencyIntervalUnit(Enum):
    # A daily billing cycle.
    DAY = 1 
    # A weekly billing cycle.
    WEEK = 2
    # A monthly billing cycle.
    MONTH = 3
    # A yearly billing cycle.
    YEAR = 4

class SetupFeeFailAction(Enum):
    # Cancels the subscription if the initial payment for the setup fails.
    CANCEL = 1
    # Continues the subscription if the initial payment for the setup fails.
    CONTINUE = 2

class PricingScheme(PayPalEntity):
    """Billing cycle pricing scheme obj representation
    """
    def __init__(self, version: int = None, fixed_price: Money = None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.version = version
        self.fixed_price = fixed_price
        self._update_time = self._json_response.get('update_time', kwargs.get('update_time'))
        self._create_time = self._json_response.get('create_time', kwargs.get('create_time'))
    
    @property
    def update_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._update_time) if self._update_time else None
        except:
            return None

    @property
    def create_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._create_time) if self._create_time else None
        except:
            return None

    @classmethod
    def create(cls, *, version: int = None, fixed_price: Money = None) -> 'PricingScheme':
        return cls(version = version, fixed_price = fixed_price)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = { **json_data }
        if 'fixed_price' in json_data.keys():
            args['fixed_price'] = Money.serialize_from_json(json_data['fixed_price'])
        return cls(**args, json_response= json_data, response_type = response_type)

class Frequency(PayPalEntity):
    """Billing cycle frequency obj representation
    """

    def __init__(self, interval_unit: str, interval_count: int = None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.interval_unit = interval_unit
        self.interval_count = interval_count

    @property
    def interval_unit_enum(self) -> FrequencyIntervalUnit:
        try:
            return FrequencyIntervalUnit[self.interval_unit] if self.interval_unit else None
        except:
            return None

    @classmethod
    def create(cls, *, interval_unit: FrequencyIntervalUnit, interval_count: int = 1) -> 'Frequency':
        return cls(interval_unit = interval_unit.name, interval_count = interval_count)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(**json_data, json_response = json_data, response_type = response_type)

class BillingCycle(PayPalEntity):
    """Plan billing cycle obj representation.
    """

    # Serializable entity types
    _ENTITY_TYPES = { 'frequency': Frequency, 'pricing_scheme': PricingScheme }

    def __init__(
            self, pricing_scheme: PricingScheme, frequency: Frequency, 
            tenure_type: str, sequence: int, total_cycles: int, **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.sequence = sequence
        self.frequency = frequency
        self.tenure_type = tenure_type
        self.total_cycles = total_cycles
        self.pricing_scheme = pricing_scheme
    
    @property
    def tenure_type_enum(self) -> BillingCycleTenureType:
        try:
            return BillingCycleTenureType[self.tenure_type] if self.tenure_type else None
        except:
            return None
    
    @classmethod
    def create(
            cls, *, frequency: Frequency, tenure_type: BillingCycleTenureType, 
            sequence: int, pricing_scheme: PricingScheme = None, total_cycles: int = 1
        ) -> 'BillingCycle':
        return cls(pricing_scheme, frequency, tenure_type.name, sequence, total_cycles)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = super()._build_args(json_data, cls._ENTITY_TYPES)
        return cls(**args, json_response= json_data, response_type = response_type)

class PaymentPreferences(PayPalEntity):
    """Payment preference obj representation
    """

    def __init__(
            self, auto_bill_outstanding: bool, setup_fee: Money, 
            setup_fee_failure_action: str, payment_failure_threshold: int, 
            **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.setup_fee = setup_fee
        self.auto_bill_outstanding = auto_bill_outstanding
        self.setup_fee_failure_action = setup_fee_failure_action
        self.payment_failure_threshold = payment_failure_threshold
    
    @property
    def setup_fee_failure_action_enum(self) -> SetupFeeFailAction:
        try:
            return SetupFeeFailAction[self.setup_fee_failure_action] if self.setup_fee_failure_action else None
        except:
            return None
    
    @classmethod
    def create(
            cls, *, auto_bill_outstanding: bool = True, payment_failure_threshold: int = 0,
            setup_fee: Money = None, setup_fee_failure_action: SetupFeeFailAction = SetupFeeFailAction.CANCEL
        ) -> 'PaymentPreferences':
        return cls(auto_bill_outstanding, setup_fee, setup_fee_failure_action.name, payment_failure_threshold)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = { **json_data }
        if 'setup_fee' in json_data.keys():
            args['setup_fee'] = Money.serialize_from_json(json_data['setup_fee'], response_type)
        return cls( **args, json_response= json_data, response_type = response_type)

class Taxes(PayPalEntity):
    """Subscription taxes obj representation
    """

    def __init__(self, percentage: str, inclusive: bool = None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.inclusive = inclusive
        self.percentage = percentage
    
    @classmethod
    def create(cls, percentage: str, inclusive: bool = True) -> 'Taxes':
        return cls(percentage, inclusive)
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(**json_data, json_response = json_data, response_type = response_type)

class Plan(PayPalEntity):
    """Subscription Plan obj representation
    """

    _ARRAY_TYPES = { 'billing_cycles': BillingCycle }
    _ENTITY_TYPES = { 'payment_preferences': PaymentPreferences, 'taxes': Taxes }

    def __init__(
            self, plan_id: str = None, product_id: str = None, 
            name: str = None, status: str = None, description: str = None, 
            billing_cycles: List[BillingCycle] = None, payment_preferences: PaymentPreferences = None,
            taxes:Taxes  = None, quantity_supported: bool = None, **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.name = name
        self.id = plan_id
        self.taxes = taxes
        self.status = status
        self.product_id = product_id
        self.description = description
        self.billing_cycles = billing_cycles
        self.quantity_supported = quantity_supported
        self.payment_preferences = payment_preferences
        self._update_time = self._json_response.get('update_time', kwargs.get('update_time'))
        self._create_time = self._json_response.get('create_time', kwargs.get('create_time'))
        self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]

    @property
    def update_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._update_time) if self._update_time else None
        except:
            return None

    @property
    def create_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._create_time) if self._create_time else None
        except:
            return None

    @property
    def status_enum(self) -> PlanStatus:
        """Status of the plan as an enum constant
        
        Returns:
            PlanStatus -- An enumerated constant representing the plan status or None
        """
        try:
            return PlanStatus[self.status] if self.status else None
        except:
            return None

    @property
    def read_link(self) -> ActionLink:
        """Retrieves a link to read this entity details.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'self', self.links), None)
    
    @classmethod
    def create(
            cls, product_id: str, name: str, billing_cycles: List[BillingCycle],
            payment_preferences: PaymentPreferences = None, description: str = None,
            status: PlanStatus = None, taxes:Taxes  = None, quantity_supported: bool = None
        ) -> 'Plan':
        return cls(
            product_id = product_id, name = name, taxes = taxes,
            description = description, billing_cycles = billing_cycles,
            quantity_supported = quantity_supported, payment_preferences = payment_preferences,
            status = status.name if status else None
        )

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = super()._build_args(json_data, cls._ENTITY_TYPES, cls._ARRAY_TYPES)
        args['plan_id'] = args.pop('id', None)
        return cls(**args, json_response= json_data, response_type = response_type)


"""
    Module with all payout related entities 
"""

from enum import Enum
from datetime import datetime
from typing import Type, List

import dateutil.parser
from pypaypal.errors import PayPalErrorDetail

from pypaypal.entities.base import ( 
    T, 
    ActionLink, 
    PaypalName,
    PayPalEntity, 
    ResponseType, 
    PaypalPhoneDetail,
    # Money is imported as currency to follow the docs naming conv.
    Money as Currency
)

class PayoutRecipientType(Enum):
    # Unencrypted email. Value is a string of up to 127 single-byte characters
    EMAIL = 1
    # The unencrypted phone number. Not supported in sandbox mode.
    PHONE = 2
    # Encrypted PayPal account number.
    PAYPAL_ID = 3

class RecipientWallet(Enum):
    # PayPal Wallet.
    PAYPAL = 1
    # Venmo Wallet.
    VENMO = 2

class BatchStatus(Enum):
    # Payout requests was denied, not processed.
    DENIED = 1 
    # Payout requests received and will be processed soon.
    PENDING = 2 
    # Payout requests received and being processed.
    PROCESSING = 3 
    # Payout batch was processed and completed.
    SUCCESS = 4 
    # The payouts file was cancelled by the sender.
    CANCELED = 5 

class TransactionStatus(Enum):
    # Funds have been credited to the recipient’s account.
    SUCCESS = 1
    # This payout request has failed, so funds were not deducted from the sender’s account.
    FAILED = 2
    # The payout request was received and will be processed.
    PENDING = 3
    # The recipient for this payout does not have a PayPal account. 
    # A link to sign up for a PayPal account was sent to the recipient. 
    # However, if the recipient does not claim this payout within 30 days, 
    # the funds are returned to your account.
    UNCLAIMED = 4
    # The recipient has not claimed this payout, 
    # so the funds have been returned to your account.
    RETURNED = 5
    # This payout request is being reviewed and is on hold.
    ONHOLD = 6
    # This payout request has been blocked.
    BLOCKED = 7
    # This payout request was refunded.
    REFUNDED = 8
    # This payout request was reversed.
    REVERSED = 9

class SenderBatchHeader(PayPalEntity):
    """ A sender-provided payout header for a payout request.
    """

    def __init__(
        self, sender_batch_id: str, recipient_type: str, 
        email_subject: str, email_message: str, note: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.note = note
        self.email_subject = email_subject
        self.email_message = email_message
        self.recipient_type = recipient_type
        self.sender_batch_id = sender_batch_id

    @property
    def recipient_type_enum(self) -> PayoutRecipientType:
        try:
            return PayoutRecipientType[self.recipient_type] if self.recipient_type else None
        except:
            return None

    @classmethod
    def create(
        cls, sender_batch_id: str, *, email_subject: str, 
        email_message: str, recipient_type: str = None, note: str = None) -> 'SenderBatchHeader':
        return cls(sender_batch_id, recipient_type, email_subject, email_message, note)
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(
            json_data.get('sender_batch_id'), json_data.get('recipient_type'), json_data.get('email_subject'), 
            json_data.get('email_message'), json_data.get('note'), json_response= json_data, 
            response_type = response_type
        )

class PayoutErrorDetails(PayPalEntity):
    """Payout error details object representation
    """
    def __init__(self, field: str, issue: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.field = field
        self.issue = issue
        self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(json_data.get('field'), json_data.get('issue'), json_response = json_data, response_type= response_type)

class PayoutError(PayPalErrorDetail):
    """Payout error object representation
    """
    
    def __init__(self, name: str, message: str, info_link: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), name, message, info_link)
        self._payout_error_details = self._json_response.get('payout_error_details')
    
    @property
    def payout_error_details(self) -> List[PayoutErrorDetails]:
        try:
            return PayoutErrorDetails.serialize_from_json(self._payout_error_details) if self._payout_error_details else []
        except:
            return []

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict) -> T:
        return cls(
            json_data.get('name'), json_data.get('message'), json_data.get('information_link'), 
            json_response = json_data
        )

class PayoutHeader(PayPalEntity):
    """Payout header object representation
    """
    def __init__(
        self, payout_batch_id: str, batch_status: str, 
        sender_batch_header: SenderBatchHeader, errors: PayPalErrorDetail, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.errors = errors
        self.batch_status = batch_status
        self.payout_batch_id = payout_batch_id
        self.sender_batch_header = sender_batch_header
        self._time_created = self._json_response.get('time_created', kwargs.get('time_created'))

    @property
    def status_enum(self) -> BatchStatus:
        try:
            return BatchStatus[self.status] if self.status else None
        except:
            return None

    @property
    def time_created(self) -> datetime:
        try:
            return dateutil.parser.parse(self._time_created) if self._time_created else None
        except:
            return None
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        errors, sender_batch_header = None, None

        if 'errors' in json_data.keys():
            errors = PayoutError.serialize_from_json(json_data['errors'])
        if 'sender_batch_header' in json_data.keys():
            sender_batch_header = SenderBatchHeader.serialize_from_json(json_data['sender_batch_header'], response_type)
        
        return cls(
            json_data.get('payout_batch_id'), json_data.get('batch_status'), 
            sender_batch_header, errors, json_response = json_data, response_type = response_type
        )

class PayoutItemDetail(PayPalEntity):
    """Payout item detail object representation
    """

    def __init__(
        self, recipient_type: str, amount: Currency, note: str, 
        receiver: str, sender_item_id: str, recipient_name: PaypalName, 
        recipient_wallet: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.note = note
        self.amount = amount
        self.receiver = receiver
        self.recipient_type = recipient_type
        self.sender_item_id = sender_item_id
        self.recipient_name  = recipient_name
        self.recipient_wallet = recipient_wallet
    
    @property
    def recipient_wallet_enum(self) -> RecipientWallet:
        try:
            return RecipientWallet[self.recipient_wallet] if self.recipient_wallet else RecipientWallet.PAYPAL
        except:
            return RecipientWallet.PAYPAL
    
    @classmethod
    def create(
        cls, *, amount: Currency, receiver: str, 
        recipient_type: str = None, note: str = None, 
        sender_item_id: str = None, recipient_name: PaypalName = None, 
        recipient_wallet: RecipientWallet = RecipientWallet.PAYPAL) -> 'PayoutItemDetail':        
        return cls(recipient_type, amount, note, receiver, sender_item_id, recipient_name, recipient_wallet.name)
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        amount, recipient_name = None, None

        if 'amount' in json_data.keys():
            amount = Currency.serialize_from_json(json_data['amount'], response_type)
        if 'recipient_name' in json_data.keys():
            recipient_name = PaypalName.serialize_from_json(json_data['recipient_name'], response_type)

        return cls(
            json_data.get('recipient_type'), amount, json_data.get('note'), 
            json_data.get('receiver'), json_data.get('sender_item_id'), recipient_name,
            json_data.get('recipient_wallet'), json_response = json_data, response_type = response_type
        )

class CurrencyConversion(PayPalEntity):
    """Currency conversion object representation
    """
    def __init__(self, from_amount: Currency, to_amount: Currency, exchange_rate: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.to_amount = to_amount
        self.from_amount = from_amount
        self.exchange_rate = exchange_rate

    @classmethod
    def create(cls, *, from_amount: Currency, to_amount: Currency, exchange_rate: str = None) -> 'CurrencyConversion':
        return cls(from_amount, to_amount, exchange_rate)
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        from_amount, to_amount = None, None

        if 'to_amount' in json_data.keys():
            to_amount = Currency.serialize_from_json(json_data['to_amount'], response_type)
        if 'from_amount' in json_data.keys():
            from_amount = Currency.serialize_from_json(json_data['from_amount'], response_type)

        return cls(from_amount, to_amount, json_data.get('exchange_rate'))

class PayoutItem(PayPalEntity):
    """Payout item object representation
    """

    def __init__(
        self, payout_item_id: str, transaction_id: str, activity_id: str, 
        transaction_status: str, payout_item_fee: Currency, payout_batch_id: str, 
        sender_batch_id: str, payout_item: PayoutItemDetail, currency_conversion: CurrencyConversion, 
        errors: PayoutError, recipient_type: str, amount: Currency, note: str, receiver: str, 
        sender_item_id: str, recipient_wallet: str, alternate_notification_method: PaypalPhoneDetail, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.note = note
        self.errors = errors
        self.amount = amount
        self.receiver = receiver
        self.activity_id = activity_id
        self.payout_item = payout_item
        self.recipient_type = recipient_type
        self.payout_item_id = payout_item_id
        self.transaction_id = transaction_id
        self.sender_item_id = sender_item_id
        self.payout_item_fee = payout_item_fee
        self.payout_batch_id = payout_batch_id
        self.sender_batch_id = sender_batch_id
        self.recipient_wallet = recipient_wallet
        self.transaction_status = transaction_status
        self.currency_conversion = currency_conversion
        self.alternate_notification_method = alternate_notification_method
        self._time_created = self._json_response.get('time_created', kwargs.get('time_created'))
        self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]

    @property
    def time_created(self) -> datetime:
        try:
            return dateutil.parser.parse(self._time_created) if self._time_created else None
        except:
            return None
    
    @classmethod
    def create(
        cls, *, amount: Currency, receiver: str, payout_item_id: str,
        payout_batch_id: str, payout_item: PayoutItemDetail,
        transaction_id: str = None, activity_id: str = None,
        transaction_status: TransactionStatus = None, payout_item_fee: Currency = None,
        sender_batch_id: str = None,  currency_conversion: CurrencyConversion = None,
        recipient_type: PayoutRecipientType = None, note: str = None, sender_item_id: str = None,
        recipient_wallet: RecipientWallet = RecipientWallet.PAYPAL, alternate_notification_method: PaypalPhoneDetail = None) -> 'PayoutItem':

        return cls(
            payout_item_id, transaction_id,activity_id, transaction_status.name, payout_item_fee, payout_batch_id,
            sender_batch_id, payout_item, currency_conversion, None, recipient_type.name, amount, note, receiver, sender_item_id,
            recipient_wallet.name, alternate_notification_method
        )

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        errors, amount, payout_item, payout_item_fee, currency_conversion, alternate_notification_method = None, None, None, None, None, None
        
        if 'errors' in json_data.keys():
            errors = PayoutError.serialize_from_json(json_data['errors'])
        if 'amount' in json_data.keys():
            amount = Currency.serialize_from_json(json_data['amount'], response_type)
        if 'payout_item' in json_data.keys():
            payout_item = PayoutItemDetail.serialize_from_json(json_data['payout_item'], response_type)
        if 'payout_item_fee' in json_data.keys():
            payout_item_fee = Currency.serialize_from_json(json_data['payout_item_fee'], response_type)
        if 'currency_conversion' in json_data.keys():
            currency_conversion = CurrencyConversion.serialize_from_json(json_data['currency_conversion'], response_type)
        if 'alternate_notification_method' in json_data.keys():
            alternate_notification_method = PaypalPhoneDetail.serialize_from_json(json_data['alternate_notification_method'], response_type)
        
        return cls(
            json_data.get('payout_item_id'), json_data.get('transaction_id'), json_data.get('activity_id'), json_data.get('transaction_status'), payout_item_fee,
            json_data.get('payout_batch_id'), json_data.get('sender_batch_id'), payout_item, currency_conversion, errors, json_data.get('recipient_type'), 
            amount, json_data.get('note'), json_data.get('receiver'), json_data.get('sender_item_id'), json_data.get('recipient_wallet'), 
            alternate_notification_method, json_response = json_data, response_type = response_type
        )

class PagedPayout(PayPalEntity):
    """Paged payout information (Payout Header + Paged Items).
    """

    def __init__(self, batch_header: PayoutHeader, items: List[PayoutItem], **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.items = items 
        self.batch_header = batch_header
        self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]
    
    @property
    def read_link(self) -> ActionLink:
        """Retrieves a link to read this entity details.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'self', self.links), None)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        batch_header, items = None, None
        
        if 'batch_header' in json_data.keys():
            batch_header = PayoutHeader.serialize_from_json(json_data['batch_header'], response_type)
        if 'items' in json_data.keys():
            items = [ PayoutItem.serialize_from_json(json_data[x]) for x in json_data['items'] ]

        return cls(batch_header, items, json_response = json_data, response_type = response_type)
    


"""Payouts resource client for the paypal resource group.

    Resource docs & Reference: https://developer.paypal.com/docs/api/payments.payouts-batch/v1/
"""

import json

from typing import Type, TypeVar, List
from pypaypal.clients.base import ClientBase, ActionLink

from pypaypal.http import ( 
    parse_url,
    PayPalSession,
    LEGACY_LIVE_API_BASE_URL,
    LEGACY_SANDBOX_API_BASE_URL
)

from pypaypal.entities.base import ( 
    PaypalPage,
    ResponseType, 
    PaypalApiResponse, 
    PatchUpdateRequest,
    PaypalApiBulkResponse
)

from pypaypal.entities.payouts import (
    PayoutItem,
    PagedPayout,
    PayoutHeader,
    SenderBatchHeader
)

"""
    Base Resource Live URL
"""
_LIVE_RESOURCE_BASE_URL = parse_url(LEGACY_LIVE_API_BASE_URL, 'payments')

"""
    Base Resource Sandbox URL
"""
_SANDBOX_RESOURCE_BASE_URL = parse_url(LEGACY_SANDBOX_API_BASE_URL, 'payments')

T = TypeVar('T', bound = 'PayoutClient')

I = TypeVar('I', bound = 'PayoutItemClient')

class PayoutClient(ClientBase):
    """Payout resource group client class.
    """
    
    def __init__(self, url: str, session: PayPalSession):
        super().__init__(url, session)
    
    def create_batch_payout(self, sender_batch_header: SenderBatchHeader, items: List[PayoutItem], request_id: str = None) -> PaypalApiResponse[PayoutHeader]:
        """Calls the paypal API to create a batch payout to one or more recipients
        
        Arguments:
            sender_batch_header {SenderBatchHeader} -- Sender-provided payout header for a payout request.
            items {List[PayoutItem]} -- A list of individual payout items.
        
        Keyword Arguments:
            request_id {str} -- Request id for idempotence (default: {None})
        
        Returns:
            PaypalApiResponse[PayoutHeader] -- A response with the generated header.
        """
        headers = dict()
        url = self._base_url

        if request_id:
            headers['PayPal-Request-Id'] = request_id
        
        body = json.dumps({ sender_batch_header.to_dict(), items.to_dict() })
        api_response = self._session.post(url, body, headers = headers) if headers else self._session.post(url, body)

        if api_response.status_code // 100 != 2:
            return PaypalApiResponse.error(api_response)

        j_data = api_response.json().get('batch_header')

        return PaypalApiResponse.success(api_response, PayoutHeader.serialize_from_json(j_data) if j_data else None)

    def show_payout_batch_details(
        self, payout_batch_id: str, page: int = 1, page_size: int = 1000, 
        fields: str = None, total_required: bool = True) -> PaypalApiResponse[PagedPayout]:
        
        url = parse_url(self._base_url, payout_batch_id)
        params = { 'page': page,  'page_size': page_size, 'total_required': total_required }

        if fields:
            params['fields'] = fields
        
        api_response = self._session.get(url, params)

        if api_response.status_code // 100 != 2:
            return PaypalApiResponse.error(api_response)

        return PaypalApiResponse.success(api_response, PagedPayout.serialize_from_json(api_response.json()))

    @classmethod
    def for_session(cls: T, session: PayPalSession) -> T:
        """Creates a client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of Dispute client with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(parse_url(base_url, 'payouts'), session)

class PayoutItemClient(ClientBase):
    """Payout items resource group client class.
    """
    def __init__(self, url: str, session: PayPalSession):
        super().__init__(url, session)

    def show_payout_item_details(self, payout_item_id: str) -> PaypalApiResponse[PayoutItem]:
        """Calls the paypal API to show a payout item details
        
        Arguments:
            payout_item_id {str} -- Desired payout item id
        
        Returns:
            PaypalApiResponse[PayoutItem] -- Response with the item data if successful
        """        
        api_response = self._session.get(parse_url(self._base_url, payout_item_id))

        if api_response.status_code // 100 != 2:
            return PaypalApiResponse.error(api_response)

        return PaypalApiResponse.success(api_response, PayoutItem.serialize_from_json(api_response.json()))

    def cancel_unclaimed_payout_item(self, payout_item_id: str) -> PaypalApiResponse[PayoutItem]:
        """Calls the paypal API to cancel an unclaimed payout item details
        
        Arguments:
            payout_item_id {str} -- Desired payout item id
        
        Returns:
            PaypalApiResponse[PayoutItem] -- Response with the item data if successful
        """        
        api_response = self._session.post(parse_url(self._base_url, payout_item_id))

        if api_response.status_code // 100 != 2:
            return PaypalApiResponse.error(api_response)

        j_data = api_response.json()
        
        if j_data:
            PaypalApiResponse.success(api_response, PayoutItem.serialize_from_json(j_data))

        return PaypalApiResponse.success(api_response)

    @classmethod
    def for_session(cls: T, session: PayPalSession) -> I:
        """Creates a client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of Dispute client with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(parse_url(base_url, 'payouts-item'), session)
